from __future__ import annotations

import datetime as dt
from dateutil import parser as dateparser
import pycallrail.base as base
import pycallrail.callrail as crl
import typing
import logging
import typeguard

class TextMessage(base.CallRailBase):
    """
    Represents a TextMessage.
    More information: https://apidocs.callrail.com/#text-messages
    """
    
    direction: str
    content: str
    created_at: dt.datetime

    def __init__(
        self,
        direction: str,
        content: str,
        created_at: dt.datetime
    ) -> None:
        super(TextMessage, self).__init__()
        if direction not in ["incoming", "outgoing"]:
            raise ValueError("direction must be 'incoming' or 'outgoing'")
        self.direction = direction
        if created_at is None or not isinstance(created_at, dt.datetime):
            raise ValueError("created_at must be a datetime")
        self.created_at = created_at
        self.content = content

class TextMessageConversation(base.CallRailBase):
    """
    A conversation/thread of text messages.
    More information: https://apidocs.callrail.com/#text-messages
    """
    
    id: str
    company_id: str
    initial_tracker_id: str
    current_tracker_id: str
    customer_name: str
    customer_phone_number: str
    initial_tracking_number: str
    current_tracking_number: str
    last_message_at: dt.datetime
    state: str
    company_time_zone: str
    formatted_customer_phone_number: str
    formatted_initial_tracking_number: str
    formatted_current_tracking_number: str
    formatted_customer_name: str
    recent_messages: typing.List[TextMessage]

    # saw these fields appear in testing, but not in the docs
    tracker_name: typing.Optional[str]
    company_name: typing.Optional[str]

    # Optional User Requested Fields
    lead_status: typing.Optional[str]

    def __init__(
        self,
        api_client: crl.CallRail,
        account_id: str,
        **kwargs
    ) -> None:
        super(TextMessageConversation, self).__init__()
        self.api_client = api_client
        self.account_id = account_id

        # validate id is in kwargs
        if 'id' not in kwargs.keys():
            raise KeyError("id is a required argument")
        
        for k,v in kwargs.items():
            # validate its a valid attribute
            if k not in self.__class__.__annotations__.keys():
                raise AttributeError(f"{k} is not a valid attribute for {self.__class__.__name__}")
            else:
                setattr(self, k, v)

    @classmethod
    def from_json(
        cls,
        api_client: crl.CallRail,
        account_id: str,
        json_data: typing.Dict[str, typing.Any]
    ) -> TextMessageConversation:
        """
        Deserialize JSON to a Text Message Conversation
        """

        json_data['last_message_at'] = dateparser.parse(json_data['last_message_at'])

        json_data['recent_messages'] = [
            TextMessage(
                message['direction'],
                message['content'],
                dateparser.parse(message['created_at'])
            ) for message in json_data['recent_messages']
        ]

        return cls(
            api_client=api_client,
            account_id=account_id,
            **json_data
        )

    def archive(self, state: str = 'archived') -> None:
        """
        Archive the conversation
        """
        self.api_client._put(
            endpoint=f'a/{self.account_id}',
            path=f'text-messages/{self.id}.json',
            params={
                'conversation_id': self.id
            },
            data = {
                'state': state
            }
        )

        self.state = state


import pytest
import pytest_mock
import typing
from pycallrail.helpers import build_url

# Tests that the function returns a valid URL string when base_url and endpoint are valid strings. 
def test_happy_path_build_url() -> None:
    # Arrange
    base_url = "https://api.callrail.com/v3/"
    endpoint = "a.json"
    expected_url = "https://api.callrail.com/v3/a.json"

    # Act
    actual_url: str = build_url(base_url, endpoint)

    # Assert
    assert actual_url == expected_url

# Tests that the function handles an empty string for base_url or endpoint. 
def test_edge_case_empty_strings() -> None:
    # Arrange
    base_url: typing.Literal[''] = ""
    endpoint = "/api/v1/users"
    expected_url = "/api/v1/users"

    # Act
    actual_url: str = build_url(base_url, endpoint)

    # Assert
    assert actual_url == expected_url

# Tests that the function handles a path that starts with a leading slash. 
def test_edge_case_leading_slash() -> None:
    # Arrange
    base_url = "https://www.example.com"
    endpoint = "/api/v1/users"
    path = "/profile"
    expected_url = "https://www.example.com/api/v1/users/profile"

    # Act
    actual_url: str = build_url(base_url, endpoint, path)

    # Assert
    assert actual_url == expected_url

# Tests that the function handles query parameters in the URL. 
def test_general_behavior_query_parameters() -> None:
    # Arrange
    base_url = "https://www.example.com"
    endpoint = "/api/v1/users"
    query_params: dict[str, str] = {"sort": "name", "filter": "active"}
    expected_url = "https://www.example.com/api/v1/users?sort=name&filter=active"

    # Act
    actual_url: str = build_url(base_url, endpoint, None)
    actual_url += "?" + "&".join([f"{k}={v}" for k, v in query_params.items()])

    # Assert
    assert actual_url == expected_url

# Tests that the function handles trailing slashes correctly for endpoint and path. 
def test_edge_case_trailing_slashes() -> None:
    # Arrange
    base_url = "https://www.example.com/"
    endpoint = "/api/v1/users/"
    path = "/profile/"
    expected_url = "https://www.example.com/api/v1/users/profile/"

    # Act
    actual_url: str = build_url(base_url, endpoint, path)

    # Assert
    assert actual_url == expected_url

# Tests that the function handles special characters in the URL. 
def test_general_behavior_special_characters() -> None:
    # Arrange
    base_url = "https://www.example.com"
    endpoint = "/api/v1/users"
    path = "/profile?name=John&age=30"
    expected_url = "https://www.example.com/api/v1/users/profile?name=John&age=30"

    # Act
    actual_url: str = build_url(base_url, endpoint, path)

    # Assert
    assert actual_url == expected_url


from __future__ import with_statement, print_function, absolute_import, annotations
import requests
import typing
import collections

from pycallrail.objects.accounts import Account
from pycallrail.helpers import build_url

class CallRail(object):
    """Base class for CallRail API access"""

    def __init__(
            self, 
            api_key: str, 
            proxies: typing.Optional[collections.MutableMapping[str, str]] = None, 
            default_pagination_type: typing.Optional[str] = 'relative'
        ) -> None:
        """
        Constructor
        
        :api_key: API Key for the CallRail Account.
        :proxies: Set of proxies to use.
        """
        if api_key is None:
            raise ValueError('API key is required')
        self.BASE_URL: str = 'https://api.callrail.com/v3/'
        self.api_key: str = api_key
        self.proxies: typing.Union[collections.MutableMapping[str, str], None] = proxies
        
        self.session: requests.Session = requests.Session()
        
        if proxies is not None:
            self.session.proxies = typing.cast(collections.MutableMapping[str, str], proxies)

        self.auth_header: collections.Mapping[str, str] = {
            "Authorization": f'Token token="{self.api_key}"'
        }

        self.session.headers.update(self.auth_header)

        if default_pagination_type == 'relative':
            self.default_pagination_param: collections.MutableMapping[str, str] = {
                'relative_pagination': 'true'
            }

    
    def _relative_paginator(
            self,
            response: requests.Response,
            response_data_key: typing.Optional[str] = None
    ) -> typing.List[typing.Dict[str, typing.Any]]:
        """
        Paginate through API responses using relative pagination
        
        :response: Response object
        :response_data_key: Key to use for response data
        """

        result_bag: list[None] = []

        while True:
            try:
                result_bag.extend(response.json()[response_data_key])
            except KeyError:
                break
            if "next_page" not in response.json() \
                or "has_next_page" not in response.json() \
                    or response.json()['has_next_page'] is False:
                break
            try:
                response = self.session.get(
                    url=response.json()['next_page'],
                    params=self.default_pagination_param
                )
                response.raise_for_status()
            except Exception as e:
                raise e
        
        return typing.cast(typing.List[typing.Dict[str, typing.Any]], result_bag)

    def _offset_paginator(
            self,
            response: requests.Response,
            response_data_key: typing.Optional[str] = None
    ) -> typing.List[typing.Dict[str, typing.Any]]:
        """
        Paginate through API responses using offset pagination
        
        :response: Response object
        :response_data_key: Key to use for response data
        """

        result_bag: list[None] = []
        
        current_page: int = response.json()['page']
        total_pages: int = response.json()['total_pages']

        while True:
            try:
                result_bag.extend(response.json()[response_data_key])
            except KeyError:
                break
            if current_page == total_pages:
                break
            try:
                response = self.session.get(
                    url=f'{response.url}',
                    params={'page': current_page + 1},
                    headers=self.auth_header
                )
                response.raise_for_status()
            except Exception as e:
                raise e
            current_page = response.json()['page']

        return typing.cast(typing.List[typing.Dict[str, typing.Any]], result_bag)
    
    def _get(
            self,
            endpoint: str,
            response_data_key: typing.Optional[str] = None,
            path: typing.Optional[str] = None,
            params: typing.Optional[typing.MutableMapping[str, typing.Any]] = None,
            pagination_type: typing.Optional[str] = 'OFFSET'
    ) -> typing.Union[typing.List[typing.Dict[str, typing.Any]], typing.Dict[str, typing.Any], None]:
        """
        Make a GET request to the CallRail API.
        
        :endpoint: API endpoint
        :path: API path
        :params: Query string parameters
        :response_data_key: Key to use for response data
        """
        if path:
            url: str = build_url(
                base_url=self.BASE_URL,
                endpoint=endpoint,
                path=path
            )
        else:
            url = build_url(
                base_url=self.BASE_URL,
                endpoint=endpoint
            )

        if params:
            if pagination_type == 'RELATIVE':
                params.update(self.default_pagination_param)
            first_response: requests.Response = self.session.get(
                url=url,
                params=params
            )
        else:
            if pagination_type == 'RELATIVE':
                first_response: requests.Response = self.session.get( # type: ignore
                    url=url,
                    params=self.default_pagination_param
                )
            else:
                first_response: requests.Response = self.session.get( # type: ignore
                    url=url
                )

        first_response.raise_for_status()

        if pagination_type == 'OFFSET':
            return self._offset_paginator(
                response=first_response,
                response_data_key=response_data_key
            )
        elif pagination_type == 'RELATIVE':
            return self._relative_paginator(
                response=first_response,
                response_data_key=response_data_key
            )
        
        else:
            return first_response.json()
        
    def _post(
            self,
            endpoint: str,
            path: typing.Optional[str] = None,
            data: typing.Optional[typing.MutableMapping[str, typing.Any]] = None,
            response_data_key: typing.Optional[str] = None
    ) -> typing.Union[typing.List[typing.Dict[str, typing.Any]], typing.Dict[str, typing.Any], None]:
        """
        Make a POST request to the CallRail API.
        
        :endpoint: API endpoint
        :path: API path
        :data: JSON data to send
        :response_data_key: Key to use for response data
        """
        if path:
            url: str = build_url(
                base_url=self.BASE_URL,
                endpoint=endpoint,
                path=path
            )
        else:
            url = build_url(
                base_url=self.BASE_URL,
                endpoint=endpoint
            )

        if data:
            first_response: requests.Response = self.session.post(
                url=url,
                json=data
            )
        else:
            first_response: requests.Response = self.session.post( # type: ignore
                url=url
            )
        
        first_response.raise_for_status()
        return first_response.json()[response_data_key] if response_data_key else first_response.json()
    
    def _put(
            self,
            endpoint: str,
            path: typing.Optional[str],
            data: typing.Optional[typing.MutableMapping[str, typing.Any]],
            response_data_key: typing.Optional[str] = None,
            params: typing.Optional[typing.Mapping[str, typing.Any]] = None
    ) -> typing.Union[typing.List[typing.Dict[str, typing.Any]], typing.Dict[str, typing.Any], None]:
        """
        Make a PUT request to the CallRail API.

        :endpoint: API endpoint
        :path: API path
        :data: JSON data to send
        :response_data_key: Key to use for response data
        """
        if path:
            url: str = build_url(
                base_url=self.BASE_URL,
                endpoint=endpoint,
                path=path
            )
        else:
            url = build_url(
                base_url=self.BASE_URL,
                endpoint=endpoint
            )
        
        if data and params:
            response: requests.Response = self.session.put(
                url=url,
                json=data,
                params=params
            )
        elif data:
            response: requests.Response = self.session.put( # type: ignore
                url=url,
                json=data
            )
        elif params:
            response: requests.Response = self.session.put( # type: ignore
                url=url,
                params=params
            )
        else:
            response: requests.Response = self.session.put( # type: ignore
                url=url
            )
            
        response.raise_for_status()

        return response.json()[response_data_key] if response_data_key else response.json()
    
    def _delete(
            self,
            endpoint: str,
            path: typing.Optional[str] = None,
            response_data_key: typing.Optional[str] = None
    ) -> None:
        """
        Make a DELETE request to the CallRail API.

        endpoint: API endpoint
        :path: API path
        :response_data_key: Key to use for response data
        """
        if path:
            url: str = build_url(
                base_url=self.BASE_URL,
                endpoint=endpoint,
                path=path
            )
        else:
            url = build_url(
                base_url=self.BASE_URL,
                endpoint=endpoint
            )
            
        
        response: requests.Response = self.session.delete(
            url=url
        )
        response.raise_for_status()

    #########################
    # Accounts
    #########################

    def list_accounts(
            self,
            **kwargs
    ) -> typing.List[Account]:
        """
        List accounts for the authenticated user.
        """

        sorting: typing.Optional[collections.MutableMapping[str, typing.Any]] = kwargs.get('sorting', None)
        filtering: typing.Optional[collections.MutableMapping[str, typing.Any]] = kwargs.get('filtering', None)
        searching: typing.Optional[collections.MutableMapping[str, typing.Any]] = kwargs.get('searching', None)
        fields: typing.Optional[collections.MutableMapping[str, typing.Any]] = kwargs.get('fields', None)

        params: dict[str, typing.Any] = {}

        if sorting:
            params |= sorting
        if filtering:
            params |= filtering
        if searching:
            params |= searching
        if fields:
            params |= fields

        response: typing.List[typing.Dict[str, typing.Any]] = typing.cast(typing.List[typing.Dict[str, typing.Any]], self._get(
            endpoint='a.json',
            response_data_key='accounts',
            params=params
        ))

        return [Account.from_json(api_client=self, json_data=account) for account in response]
    
    def get_account(
            self,
            account_id: str,
            **kwargs
    ) -> Account:
        """
        Get an account by ID.
        """
        fields: typing.Optional[collections.MutableMapping[str, typing.Any]] = kwargs.get('fields', None)
        params: dict[str, typing.Any] = {}

        if fields:
            params |= fields

        return Account.from_json(
            api_client=self, 
            json_data=typing.cast(typing.Dict[str, typing.Any],self._get(
                endpoint='a',
                response_data_key='accounts',
                path=f'/{account_id}.json',
                params=params,
                pagination_type='NONE'
            ))
        )
    


import typing

class CallRailBase(object):
    def __init__(self) -> None:
        self.id: typing.Union[str, None, int] = None

    def __hash__(self) -> int:
        class_name = type(self).__name__
        return hash((class_name, self.id))


from __future__ import annotations

import datetime as dt
from dateutil import parser as dateparser
import pycallrail.base as base
import pycallrail.callrail as crl
import typing
import typing_extensions
import requests


class FormSubmission(base.CallRailBase):
    """
    Represents a form submission.

    More info: https://apidocs.callrail.com/#form-submissions
    """

    # These fields are always returned by the API
    id: str
    company_id: str
    person_id: str
    form_data: typing.Dict[str, typing.Any]
    form_url: str
    landing_page_url: str
    referrer: str
    referring_url: str
    submitted_at: dt.datetime
    first_form: bool
    customer_phone_number: str
    customer_name: str
    formatted_customer_phone_number: str
    formatted_customer_name: str
    source: str
    keywords: str
    campaign: str
    medium: str

    # User Requested Fields
    lead_status: typing.Optional[str]
    value: typing.Optional[typing.Union[int, float]]
    note: typing.Optional[str]
    tags: typing.Optional[typing.List[typing.Any]]
    utm_source: typing.Optional[str]
    utm_campaign: typing.Optional[str]
    form_name: typing.Optional[str]
    timeline_url: typing.Optional[str]
    milestones: typing.Optional[typing.Any]

    def __init__(
        self,
        api_client: crl.CallRail,
        account_id: str,
        **kwargs
    ) -> None:
        super(FormSubmission, self).__init__()
        self.api_client: crl.CallRail = api_client
        self.account_id: str = account_id

        for k,v in kwargs.items():
            if k not in self.__class__.__annotations__.keys():
                raise AttributeError(f'{k} is not a valid attribute for {self.__class__.__name__}')
            else:
                setattr(self, k, v)

    @classmethod
    def from_json(
        cls,
        api_client: crl.CallRail,
        account_id: str,
        json_data: typing.Dict[str, typing.Any]
    ) -> FormSubmission:
        """
        Deserializes a FormSubmission from JSON.

        :param api_client: The CallRail API client
        :param json_data: The JSON data to deserialize
        :param account_id: The CallRail account ID
        :return: The FormSubmission object
        """
        json_data['submitted_at'] = dateparser.parse(json_data['submitted_at'])
        
        return cls(
            api_client=api_client,
            account_id=account_id,
            **json_data
        )
        
    def update(
        self,
        tags: typing.Optional[typing.List[typing.Any]] = None,
        note: typing.Optional[str] = None,
        value: typing.Optional[typing.Union[int, float]] = None,
        lead_status: typing.Optional[str] = None,
        append_tags: typing.Optional[bool] = None
    ) -> None:
        """
        Update a Form Submission.
        More information: https://apidocs.callrail.com/#updating-a-form-submission
        """

        # update the attributes
        if append_tags and tags and hasattr(self, 'tags'):
            self.tags.extend(tags) # type: ignore
        elif tags:
            self.tags = tags
        if note:
            self.note = note
        if value:
            self.value = value
        if lead_status:
            self.lead_status = lead_status

        # form the request
        body: typing.Dict[str, typing.Any] = {
            'tags': tags,
            'note': note,
            'value': value,
            'lead_status': lead_status
        }

        # send the request
        response: typing.Dict[str, typing.Any] = typing.cast(
            typing.Dict[str, typing.Any],
            self.api_client._put(
                endpoint=f'a/{self.account_id}',
                path=f'form_submissions/{self.id}.json',
                data=body
            )
        )

        # update the object
        for k,v in response.items():
            # raise an error if the attribute is not found
            if k not in self.__class__.__annotations__.keys():
                raise AttributeError(f'{k} is not a valid attribute for {self.__class__.__name__}')
            else:
                setattr(self, k, v)

        


from __future__ import annotations

import datetime as dt
from dateutil import parser as dateparser
import pycallrail.base as base
import pycallrail.callrail as crl
import typing
import typing_extensions
import logging
import typeguard
from pycallrail.errors import LightValidationError

@typeguard.typechecked
class Company(base.CallRailBase):
    """
    Represents a Company.

    More information: https://apidocs.callrail.com/#companies
    """

    id: str
    name: str
    status: str
    time_zone: str
    created_at: dt.datetime
    disabled_at: typing.Optional[dt.datetime]
    dni_active: typing.Optional[bool]
    script_url: str
    callscore_enabled: bool
    lead_scoring_enabled: bool
    
    
    swap_exclude_jquery: typing.Optional[bool]
    """ Deprecated """
    swap_ppc_override: typing.Optional[bool]
    swap_landing_override: typing.Optional[str]
    swap_cookie_duration: typing.Optional[int]

    callscribe_enabled: bool
    keyword_spotting_enabled: typing.Optional[bool]
    """ Deprecated """
    form_capture: bool

    def __init__(
        self,
        api_client: crl.CallRail,
        account_id: str,
        **kwargs
    ) -> None:
        super(Company, self).__init__()
        self.api_client: crl.CallRail = api_client
        self.account_id = account_id

        if not kwargs:
            raise TypeError(f'{self.__class__.__name__}() missing data arguments! Whats going on?')
        
        for k,v in kwargs.items():
            if k not in self.__class__.__annotations__.keys():
                raise AttributeError(f'{k} is not a valid attribute for {self.__class__.__name__}')
            else:
                setattr(self, k, v)

    @classmethod
    def from_json(
        cls,
        api_client: crl.CallRail,
        account_id: str,
        json_data: typing.Dict[str, typing.Any]
    ) -> Company:
        """
        Deserialize a Company from a JSON object.
        
        :param api_client: The CallRail API client.
        :param account_id: The account ID.
        :param json_data: The JSON data.

        :return: A Company object.
        """

        json_data['disabled_at'] = dateparser.parse(json_data['disabled_at']) if json_data['disabled_at'] else None

        return cls(api_client, account_id, **json_data)

    def delete(self) -> None:
        """
        Delete the Company.
        """

        self.api_client._delete(
            endpoint=f'/a/{self.account_id}',
            path=f'/companies/{self.id}.json'
        )

    def update(
            self,
            **kwargs
    ) -> None:
        
        UPDATABLE_FIELDS: typing.List[str] = [
            'name',
            'callscore_enabled',
            'keyword_spotting_enabled',
            'callscribe_enabled',
            'time_zone',
            'swap_exclude_jquery',
            'swap_ppc_override',
            'swap_landing_override',
            'swap_cookie_duration',
            'external_form_capture'
        ]

        body = {}

        for k,v in kwargs.items():
            if k not in UPDATABLE_FIELDS:
                raise LightValidationError(f'{k} is not updatable!')
            else:
                setattr(self, k, v)
                body[k] = v

        self.api_client._put(
            endpoint=f'/a/{self.account_id}',
            path=f'/companies/{self.id}.json',
            data=body
        )


"""
CallRail API Wrapper
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

A Python wrapper for the CallRail API.

:copyright: (c) 2023 by Predictive Data Lab
:license: MIT, see LICENSE for more details.

"""

__title__ = 'pycallrail'
__author__ = 'Engineering @ Predictive Data Lab'
__license__ = 'MIT'
__copyright__ = 'Copyright 2023 by Predictive Data Lab'
__version__ = '0.9.0.1'

__path__ = __import__('pkgutil').extend_path(__path__, __name__)

import logging
from typing import NamedTuple, Literal

from callrail import *
from base import *
from errors import *
from helpers import *
from mixins import *
from objects import *

class VersionInfo(NamedTuple):
    major: int
    minor: int
    revision: int
    build: int
    release: Literal['alpha', 'beta', 'rc', 'final']

version_info: VersionInfo = VersionInfo(
    major=0,
    minor=9,
    revision=0,
    build=0,
    release='final'
)

logging.getLogger(__name__).addHandler(logging.NullHandler())

del logging, NamedTuple, Literal, VersionInfo


import pytest
import pytest_mock
import requests_mock
import requests

from pycallrail.callrail import CallRail
from pycallrail.objects.textmessages import TextMessage, TextMessageConversation
from pycallrail.objects.accounts import Account
from pycallrail.objects.companies import Company
from pycallrail.objects.calls import Call
import typing
import logging
import datetime as dt
import typeguard
import dataclasses

# Tests that calls can be successNoney listed with various keyword arguments. 
def test_list_calls(mocker: pytest_mock.MockerFixture) -> None:
    # Arrange
    api_client = CallRail('test_key')
    account = Account(api_client, 'test_id', 'test_name', True, False)
    mocker.patch.object(api_client, '_get', return_value=[
        {
        "answered": False,
        "business_phone_number": None,
        "customer_city": "Denver",
        "customer_country": "US",
        "customer_name": "RUEGSEGGER SIMO",
        "customer_phone_number": "+13036231131",
        "customer_state": "CO",
        "direction": "inbound",
        "duration": 60,
        "id": "CAL8154748ae6bd4e278a7cddd38a662f4f",
        "recording": "https://api.callrail.com/v3/a/227799611/calls/111222333/recording.json",
        "recording_duration": "27",
        "recording_player": "https://app.callrail.com/calls/111222333/recording?access_key=3b91eb7f7cc08a4d01ed",
        "start_time": "2017-01-24T11:27:48.119-05:00",
        "tracking_phone_number": "+13038163491",
        "voicemail": False,
        "agent_email": "gil@televised.com"
        },
        {
        "answered": False,
        "business_phone_number": None,
        "customer_city": "Blue Ridge",
        "customer_country": "US",
        "customer_name": "BLUE RIDGE, GA",
        "customer_phone_number": "+17064558047",
        "customer_state": "GA",
        "direction": "inbound",
        "duration": 120,
        "id": "CAL8154748ae6bd4e278a7cddd38a662f4f",
        "recording": None,
        "recording_duration": None,
        "recording_player": None,
        "start_time": "2017-01-24T19:50:03.456-05:00",
        "tracking_phone_number": "+17708243899",
        "voicemail": False,
        "agent_email": "elbert@bpp.com"
        }
    ])

    # Act
    calls: typing.List[Call] = typing.cast(typing.List[Call],
            account.list_calls(pagination_type='RELATIVE', sorting='duration', filtering={'status': 'completed'})
    )

    # Assert
    assert len(calls) == 2
    assert calls[0].id == 'CAL8154748ae6bd4e278a7cddd38a662f4f'
    assert calls[0].duration == 60
    assert calls[1].id == 'CAL8154748ae6bd4e278a7cddd38a662f4f'
    assert calls[1].duration == 120

# Tests that a single call can be successNoney retrieved with optional field selection. 
def test_get_call(mocker: pytest_mock.MockerFixture) -> None:
    # Arrange
    api_client = CallRail('test_key')
    account = Account(api_client, 'test_id', 'test_name', True, False)
    mocker.patch.object(api_client, '_get', return_value={
    "answered": False,
    "business_phone_number": None,
    "customer_city": "New York City",
    "customer_country": "US",
    "customer_name": "Jimmy Pesto, Sr.",
    "customer_phone_number": "+13036231131",
    "customer_state": "NY",
    "direction": "inbound",
    "duration": 4,
    "id": "CAL8154748ae6bd4e278a7cddd38a662f4f",
    "recording": "https://api.callrail.com/v3/a/227799611/calls/111222333/recording.json",
    "recording_duration": "27",
    "recording_player": "https://app.callrail.com/calls/111222333/recording?access_key=3b91eb7f7cc08a4d01ed",
    "start_time": "2017-01-24T11:27:48.119-05:00",
    "tracking_phone_number": "+13038163491",
    "voicemail": False,
    "agent_email": "gil@televised.com"
    })

    # Act
    call = account.get_call('1')

    # Assert
    assert call.id == 'CAL8154748ae6bd4e278a7cddd38a662f4f'
    assert call.duration == 4

# Tests that a call can be successNoney created with various optional parameters. 
def test_create_call(mocker: pytest_mock.MockerFixture) -> None:
    # Arrange
    api_client = CallRail('test_key')
    account = Account(api_client, 'test_id', 'test_name', True, False)
    mocker.patch.object(api_client, '_post', return_value={
        'id': '1', 
        'duration': 60, 
        'start_time': '2017-01-24T11:27:48.119-05:00'})


    # Act
    call: Call = account.create_call(1234567890, '555-555-5555', '444-444-4444', recording_enabled=True)

    # Assert
    assert call.id == '1'
    assert call.duration == 60

# Tests that a tag can be successNoney created with various optional parameters, and that missing required parameters raise a ValueError. 
def test_create_tag(mocker: pytest_mock.MockerFixture) -> None:
    # Arrange
    api_client = CallRail('test_key')
    account = Account(api_client, 'test_id', 'test_name', True, False)
    mocker.patch.object(api_client, '_post', return_value={
    "id": "1234569",
    "name": "Existing Customer",
    "tag_level": "company",
    "color": "gray1",
    "background_color": "gray1",
    "company_id": "COM8154748ae6bd4e278a7cddd38a662f4f",
    "status": "enabled",
    "created_at": "2014-06-06T12:11:02.964-04:00"
    })

    # Act
    tag = account.create_tag(
        name='Existing Customer',
        company_id='COM8154748ae6bd4e278a7cddd38a662f4f',
        tag_level='company',
        color='gray1'
    )

    # Assert
    assert tag.id == '1234569'
    assert tag.name == 'Existing Customer'


    # Act/Assert - Test missing required parameter
    with pytest.raises(ValueError):
        account.create_tag(None)

# Tests that companies can be successNoney listed with various keyword arguments, and that invalid filtering fields raise a ValueError. 
def test_list_companies(mocker: pytest_mock.MockerFixture) -> None:
    # Arrange
    api_client = CallRail('test_key')
    account = Account(api_client, 'test_id', 'test_name', True, False)
    returnval = [
        {
            "id": "COM8154748ae6bd4e278a7cddd38a662f4f",
            "name": "Widget Shop",
            "status": "active",
            "time_zone": "America/New_York",
            "created_at": "2016-06-06T07:06:01.000Z",
            "disabled_at": None,
            "dni_active": True,
            "script_url": "//cdn.callrail.com/companies/183570817/5706dbe6dc972c634a65/12/swap.js",
            "callscribe_enabled": False,
            "lead_scoring_enabled": True,
            "swap_exclude_jquery": None,
            "swap_ppc_override": None,
            "swap_landing_override": None,
            "swap_cookie_duration": 365,
            "callscore_enabled": False,
            "keyword_spotting_enabled": False,
            "form_capture": True
    },
    {
        "id": "COM8154748ae6bd4e278a7cddd38a662f4f",
        "name": "Bob's Burger",
        "status": "active",
        "time_zone": "America/New_York",
        "created_at": "2017-06-20T09:30:08.676Z",
        "disabled_at": None,
        "dni_active": False,
        "script_url": "//cdn.callrail.com/companies/411892629/1dbf0fc11eabdc04483a/12/swap.js",
        "callscribe_enabled": False,
        "lead_scoring_enabled": True,
        "swap_exclude_jquery": None,
        "swap_ppc_override": None,
        "swap_landing_override": None,
        "swap_cookie_duration": 365,
        "callscore_enabled": False,
        "keyword_spotting_enabled": False,
        "form_capture": True
    },
    {
        "id": "COM8154748ae6bd4e278a7cddd38a662f4f",
        "name": "Joe's Icecream",
        "status": "active",
        "time_zone": "America/New_York",
        "created_at": "2016-06-06T07:06:01.000Z",
        "disabled_at": None,
        "dni_active": None,
        "script_url": "//cdn.callrail.com/companies/785622206/81fbf538633804b8dea6/12/swap.js",
        "callscribe_enabled": False,
        "lead_scoring_enabled": True,
        "swap_exclude_jquery": None,
        "swap_ppc_override": None,
        "swap_landing_override": None,
        "swap_cookie_duration": 365,
        "callscore_enabled": False,
        "keyword_spotting_enabled": False,
        "form_capture": False
    }]
    mocker.patch.object(api_client, '_get', return_value=returnval)

    companies = account.list_companies(
        pagination_type='OFFSET', 
        sorting='name', 
        filtering={'status': 'active'})

    # Assert
    assert len(companies) == 3
    assert companies[0].id == 'COM8154748ae6bd4e278a7cddd38a662f4f'
    assert companies[0].name == 'Widget Shop'
    assert companies[1].id == 'COM8154748ae6bd4e278a7cddd38a662f4f'
    assert companies[1].name == 'Bob\'s Burger'
    assert companies[2].id == 'COM8154748ae6bd4e278a7cddd38a662f4f'
    assert companies[2].name == 'Joe\'s Icecream'

    # Act/Assert - Test invalid filtering field
    with pytest.raises(ValueError):
        account.list_companies(filtering={'invalid_field': 'value'})

# Tests that tags can be successNoney listed with various keyword arguments. 
def test_list_tags(mocker: pytest_mock.MockerFixture) -> None:
    # Arrange
    api_client = CallRail('test_key')
    account = Account(api_client, 'test_id', 'test_name', True, False)
    return_val = [
    {
      "id": 993186,
      "name": "Conversion",
      "tag_level": "company",
      "color": "gray1",
      "background_color": "green2",
      "company_id": "COM37221d54e80c4216898d2f857fc69fa0",
      "status": "enabled",
      "created_at": "2017-07-24T11:13:47.366-04:00"
    },
    {
      "id": 1932597,
      "name": "Sales",
      "tag_level": "company",
      "color": "gray1",
      "background_color": "blue2",
      "company_id": "COM37221d54e80c4216898d2f857fc69fa0",
      "status": "enabled",
      "created_at": "2019-04-05T13:38:36.258-04:00"
    },
    {
      "id": 2672419,
      "name": "Unanswered",
      "tag_level": "company",
      "color": "gray1",
      "background_color": "yellow1",
      "company_id": "COM37221d54e80c4216898d2f857fc69fa0",
      "status": "enabled",
      "created_at": "2020-10-16T16:07:35.554-04:00"
    }
  ]
    mocker.patch.object(api_client, '_get', return_value=return_val)

    # Act
    tags = account.list_tags(pagination_type='ABSOLUTE', sorting='name', filtering={'status': 'active'})

    # Assert
    assert len(tags) == 3
    assert tags[0].id == 993186
    assert tags[0].name == 'Conversion'
    assert tags[1].id == 1932597
    assert tags[1].name == 'Sales'
    assert tags[2].id == 2672419
    assert tags[2].name == 'Unanswered'

# Tests that a single company can be successNoney retrieved with optional field selection.  
def test_get_company(mocker: pytest_mock.MockerFixture):
    # Arrange
    api_client = CallRail('test_key')
    account = Account(api_client, 'test_id', 'test_name', True, False)
    company_id = 'COM8154748ae6bd4e278a7cddd38a662f4f'
    expected_url = f'a/{account.id}'
    expected_path = f'companies/{company_id}.json'
    expected_data = {
        "id": "COM8154748ae6bd4e278a7cddd38a662f4f",
        "name": "Widget Shop",
        "status": "active",
        "time_zone": "America/New_York",
        "created_at": "2016-06-06T07:06:01.000Z",
        "disabled_at": None,
        "dni_active": True,
        "script_url": "//cdn.callrail.com/companies/183570817/5706dbe6dc972c634a65/12/swap.js",
        "callscore_enabled": False,
        "lead_scoring_enabled": True,
        "swap_exclude_jquery": None,
        "swap_ppc_override": None,
        "swap_landing_override": None,
        "swap_cookie_duration": 365,
        "callscribe_enabled": False,
        "keyword_spotting_enabled": False,
        "form_capture": True,
    }
    mocker.patch.object(api_client, '_get', return_value=expected_data)

    # Act
    company: Company = account.get_company(company_id)

    # Assert
    api_client._get.assert_called_once_with(endpoint=expected_url, path=expected_path)
    assert isinstance(company, Company)
    assert company.id == company_id
    assert company.name == 'Widget Shop'

# Tests that a company can be successNoney created with various optional parameters.  
def test_create_company(mocker: pytest_mock.MockerFixture) -> None:
    # Arrange
    api_client = CallRail('test_key')
    account = Account(api_client, 'test_id', 'test_name', True, False)
    expected_url = f'a/{account.id}'
    expected_path = f'companies.json'
    
    expected_data = {
        "id": "COM8154748ae6bd4e278a7cddd38a662f4f",
        "name": "Widget Shop",
        "status": "active",
        "time_zone": "America/New_York",
        "created_at": "2017-01-26T22:58:04.184Z",
        "disabled_at": None,
        "dni_active": None,
        "script_url": "//cdn.callrail.com/companies/279054151/a74c824140d67442debd/12/swap.js",
        "callscribe_enabled": False,
        "lead_scoring_enabled": False,
        "swap_exclude_jquery": None,
        "swap_ppc_override": None,
        "swap_landing_override": None,
        "swap_cookie_duration": 365,
        "callscore_enabled": False,
        "keyword_spotting_enabled": False
    }

    mocker.patch.object(api_client, '_post', return_value=expected_data)

    # Act
    company = account.create_company(name='Widget Shop', time_zone='America/New_York')

    # Assert
    api_client._post.assert_called_once_with(endpoint=expected_url, path=expected_path, data={'name': 'Widget Shop', 'time_zone': 'America/New_York'})
    assert isinstance(company, Company)
    assert company.id == 'COM8154748ae6bd4e278a7cddd38a662f4f'
    assert company.name == 'Widget Shop'

    # Tests that text message conversations can be successfully listed with various keyword arguments.  
def test_list_text_message_conversations(mocker: pytest_mock.MockerFixture) -> None:
    # Arrange
    api_client = CallRail('test_key')
    account = Account(api_client, 'test_id', 'test_name', True, False)
    expected_url = f'a/{account.id}'
    expected_path = f'text-messages.json'
    expected_data = [
    {
      "id": "KZaGR",
      "company_id": "COM8154748ae6bd4e278a7cddd38a662f4f",
      "initial_tracker_id": "TRK8154748ae6bd4e278a7cddd38a662f4f",
      "current_tracker_id": "TRK8154748ae6bd4e278a7cddd38a662f5f",
      "customer_name": "MANN KEVIN",
      "customer_phone_number": "+14042223333",
      "initial_tracking_number": "+14045556677",
      "current_tracking_number": "+17703334455",
      "last_message_at": "2016-07-28T19:26:43.578Z",
      "state": "active",
      "company_time_zone": "America/New_York",
      "formatted_customer_phone_number": "404-222-3333",
      "formatted_initial_tracking_number": "404-555-6677",
      "formatted_current_tracking_number": "770-333-4455",
      "formatted_customer_name": "Mann Kevin",
      "recent_messages": [
        {
          "direction": "outgoing",
          "content": "Awww! But I was going into Tosche Station to pick up some power converters!",
          "created_at": "2016-07-28T19:28:21.578Z"
        },
        {
          "direction": "incoming",
          "content": "Take these two over to the garage, will you?  I want them cleaned up before dinner.",
          "created_at": "2016-07-28T19:26:43.578Z"
        }
      ]
    },
    {
      "id": "KZaGY",
      "company_id": "COM8154748ae6bd4e278a7cddd38a662f4f",
      "initial_tracker_id": "TRK8154748ae6bd4e278a7cddd38a662f4f",
      "current_tracker_id": "TRK8154748ae6bd4e278a7cddd38a662f4f",
      "customer_name": "POWELL ANDY",
      "customer_phone_number": "+14042223333",
      "initial_tracking_number": "+14045556677",
      "current_tracking_number": "+17703334455",
      "last_message_at": "2016-07-28T19:26:43.578Z",
      "state": "active",
      "company_time_zone": "America/New_York",
      "formatted_customer_phone_number": "404-222-3333",
      "formatted_initial_tracking_number": "404-555-6677",
      "formatted_current_tracking_number": "770-333-4455",
      "formatted_customer_name": "Powell Andy",
      "recent_messages": [
        {
          "direction": "outgoing",
          "content": "It’s not impossible. I used to bullseye womp rats in my T-16 back home, they’re not much bigger than two meters.",
          "created_at": "2016-07-28T19:28:21.578Z"
        },
        {
          "direction": "incoming",
          "content": "That's impossible, even for a computer!",
          "created_at": "2016-07-28T19:26:43.578Z"
        }
      ]
    }
  ]


    mocker.patch.object(api_client, '_get', return_value=expected_data)

    # Act
    conversations = account.list_text_message_conversations()

    # Assert
    api_client._get.assert_called_once_with(
        endpoint=expected_url, 
        path=expected_path, 
        params=None, 
        response_data_key='conversations',
        pagination_type='RELATIVE'
    )

    assert isinstance(conversations, list)
    assert len(conversations) == 2
    assert isinstance(conversations[0], TextMessageConversation)
    assert isinstance(conversations[1], TextMessageConversation)
    assert conversations[0].id == 'KZaGR'
    assert conversations[1].id == 'KZaGY'
    assert conversations[0].company_id == 'COM8154748ae6bd4e278a7cddd38a662f4f'
    assert conversations[1].company_id == 'COM8154748ae6bd4e278a7cddd38a662f4f'
    assert conversations[0].initial_tracker_id == 'TRK8154748ae6bd4e278a7cddd38a662f4f'
    assert conversations[1].initial_tracker_id == 'TRK8154748ae6bd4e278a7cddd38a662f4f'

    # assert text messages were created
    assert isinstance(conversations[0].recent_messages[0], TextMessage)
    assert conversations[0].recent_messages[0].content == 'Awww! But I was going into Tosche Station to pick up some power converters!'
    assert isinstance(conversations[0].recent_messages[1], TextMessage)
    assert conversations[0].recent_messages[1].content == 'Take these two over to the garage, will you?  I want them cleaned up before dinner.'


import pytest
import pytest_mock

import os
import typing
from pycallrail.callrail import CallRail
from pycallrail.objects.accounts import Account
from pycallrail.objects.calls import Call
from pycallrail.objects.companies import Company
from pycallrail.objects.form_submissions import FormSubmission
from pycallrail.objects.textmessages import TextMessageConversation
import logging


@pytest.fixture(scope='module')
def callrail() -> typing.Generator[CallRail, None, None]:
    if os.getenv('CALLRAIL_API_KEY'):
        api_key: str = os.getenv('CALLRAIL_API_KEY')
    else:
        from . import cfg
        api_key = cfg.CALLRAIL_API_KEY

    yield CallRail(api_key)

@pytest.mark.integration
def test_list_accounts(callrail: CallRail) -> None:
    
    accounts = callrail.list_accounts()


    assert isinstance(accounts, list)
    assert isinstance(accounts[0], Account)
    
    # assert there are attributes in the response
    assert hasattr(accounts[0], 'id')
    assert hasattr(accounts[0], 'name')

@pytest.mark.integration
def test_list_calls(callrail: CallRail) -> None:

    accounts: typing.List[Account] = callrail.list_accounts()
    account: Account = accounts[0]

    calls: typing.List[Call] = account.list_calls()

    assert isinstance(calls, list)
    assert isinstance(calls[0], Call)

    # assert there are attributes in the response
    assert hasattr(calls[0], 'id')
    assert hasattr(calls[0], 'account_id')


@pytest.mark.integration
def test_list_companies(callrail: CallRail) -> None:

    accounts = callrail.list_accounts()
    account: Account = accounts[0]
    
    companies: typing.List[Account] = account.list_companies()

    assert isinstance(companies, list)
    assert isinstance(companies[0], Company)

    # assert there are attributes in the response
    assert hasattr(companies[0], 'id')
    assert hasattr(companies[0], 'name')

@pytest.mark.integration
def test_formsubmission(callrail: CallRail) -> None:

    accounts = callrail.list_accounts()
    account: Account = accounts[0]
    
    form_submissions: typing.List[FormSubmission] = account.list_form_submissions()
    
    assert isinstance(form_submissions, list)
    assert isinstance(form_submissions[0], FormSubmission)
    
    # assert there are attributes in the response
    assert hasattr(form_submissions[0], 'id')
    assert hasattr(form_submissions[0], 'account_id')
    assert hasattr(form_submissions[0], 'form_data')

@pytest.mark.integration
def test_list_text_messages(callrail: CallRail) -> None:

    accounts = callrail.list_accounts()
    account: Account = accounts[0]
    
    text_messages: typing.List[TextMessageConversation] = account.list_text_message_conversations()
    
    assert isinstance(text_messages, list)
    assert isinstance(text_messages[0], TextMessageConversation)
    
    # assert there are attributes in the response
    assert hasattr(text_messages[0], 'id')
    assert hasattr(text_messages[0], 'account_id')
    assert hasattr(text_messages[0], 'recent_messages')


import pytest
import pytest_mock
import requests_mock
import requests

from pycallrail.callrail import CallRail
from pycallrail.objects.tags import Tag
import typing
import logging
import datetime as dt
import typeguard

# Tests creating a Tag object with valid input parameters. 
def test_create_tag_valid_input(mocker: pytest_mock.MockerFixture):
    # Arrange
    api_client = mocker.Mock(spec=CallRail)
    
    account_id = "123ABC"

    json_data: typing.Dict[str, typing.Any] = {
        "id": 1234569,
        "name": "Existing Customer",
        "tag_level": "company",
        "color": "gray1",
        "background_color": "gray1",
        "company_id": "COM8154748ae6bd4e278a7cddd38a662f4f",
        "status": "enabled",
        "created_at": "2014-06-06T12:11:02.964-04:00"
    }

    api_client._post.return_value = json_data

    # Act
    tag = Tag.from_json(api_client, account_id, json_data)

    # Assert
    assert tag.id == json_data["id"]
    assert tag.name == json_data["name"]
    assert tag.tag_level == json_data["tag_level"]
    assert tag.color == json_data["color"]
    assert tag.background_color == json_data["background_color"]
    assert tag.company_id == json_data["company_id"]
    assert tag.status ==   json_data["status"]
    assert tag.created_at == json_data["created_at"]
    assert isinstance(tag.created_at, dt.datetime)

# Tests updating a Tag object with valid input parameters. 
def test_update_tag_valid_input(mocker: pytest_mock.MockerFixture) -> None:
    # Arrange
    api_client = mocker.Mock(spec=CallRail)

    account_id = '123ABC'
    id = 1234569
    name = 'Existing Customer - Old Name'
    tag_level = 'company'
    color = 'gray2'
    background_color = 'gray1'
    company_id = 'COM8154748ae6bd4e278a7cddd38a662f4f'
    status = 'enabled'
    created_at: dt.datetime = dt.datetime.now()

    update_data: typing.Dict[str, str] = {
           "name": "Existing Customer",
           "color": "gray1"
    }

    json_data: typing.Dict[str, str] = {
        "id": "1234569",
        "name": "Existing Customer",
        "tag_level": "company",
        "color": "gray1",
        "background_color": "gray1",
        "company_id": "COM8154748ae6bd4e278a7cddd38a662f4f",
        "status": "enabled",
        "created_at": "2014-06-06T12:11:02.964-04:00"
    }

    api_client._put.return_value = json_data

    tag = Tag(api_client, account_id, id, name, tag_level, color, background_color, company_id, status, created_at)

    # Act
    tag.update(name=update_data["name"], color=update_data["color"])

    # Assert
    assert tag.name == update_data["name"]
    assert tag.color == update_data["color"]
    
# Tests creating a Tag object with invalid input parameters. 
def test_create_tag_invalid_input(mocker: pytest_mock.MockerFixture) -> None:
    # Arrange
    api_client = mocker.Mock(spec=CallRail)
    account_id = '123'
    id = 1
    name = None
    tag_level = 'company'
    color = '#FFFFFF'
    background_color = '#000000'
    company_id = '456'
    status = 'active'
    created_at: dt.datetime = dt.datetime.now()

    json_data: typing.Dict[str, typing.Any] = {
        'id': id,
        'name': name,
        'tag_level': tag_level,
        'color': color,
        'background_color': background_color,
        'company_id': company_id,
        'status': status,
        'created_at': created_at.isoformat()
    }

    api_client._post.return_value = json_data

    # Act and Assert
    with pytest.raises(typeguard.TypeCheckError):
        Tag.from_json(api_client, account_id, json_data)

# Tests deleting a Tag object. 
def test_delete_tag(mocker: pytest_mock.MockerFixture):
    # Arrange
    api_client = mocker.Mock()
    account_id = '123'
    id = 1
    name = 'Test Tag'
    tag_level = 'company'
    color = '#FFFFFF'
    background_color = '#000000'
    company_id = '456'
    status = 'active'
    created_at: dt.datetime = dt.datetime.now()

    tag = Tag(api_client, account_id, id, name, tag_level, color, background_color, company_id, status, created_at)

    # Act
    tag.delete()

    # Assert
    api_client._delete.assert_called_once_with(endpoint=f'a/{account_id}', path=f'tags/{id}.json')

# Tests deleting a non-existent Tag object. 
def test_delete_nonexistent_tag(mocker: pytest_mock.MockerFixture):
    # Arrange
    api_client = mocker.Mock()
    account_id = '123'
    id = 1
    name = 'Test Tag'
    tag_level = 'company'
    color = '#FFFFFF'
    background_color = '#000000'
    company_id = '456'
    status = 'active'
    created_at = dt.datetime.now()

    tag = Tag(api_client, account_id, id, name, tag_level, color, background_color, company_id, status, created_at)

    api_client._delete.side_effect = Exception('Tag not found')

    # Act and Assert
    with pytest.raises(Exception):
        tag.delete()


from __future__ import annotations

import datetime as dt
from dateutil import parser as dateparser
import pycallrail.base as base
import pycallrail.callrail as crl
import typing
import logging
import typeguard

@typeguard.typechecked
class Tag(base.CallRailBase):
    """
    Represents a Tag.

    More information: https://apidocs.callrail.com/#tags
    """

    id: typing.Union[int, str]
    name: str
    tag_level: str
    color: str
    background_color: str
    company_id: str
    status: str
    created_at: dt.datetime

    def __init__(
        self,
        api_client: crl.CallRail,
        account_id: str,
        id: typing.Union[int, str],
        name: str,
        tag_level: str,
        color: str,
        background_color: str,
        company_id: str,
        status: str,
        created_at: dt.datetime
    ) -> None:
        super(Tag, self).__init__()
        self.api_client: crl.CallRail = api_client
        self.account_id = account_id
        self.id = id
        self.name = name    
        self.tag_level = tag_level
        self.color = color
        self.background_color = background_color
        self.company_id = company_id
        self.status = status
        self.created_at = created_at

    @classmethod
    def from_json(
        cls,
        api_client: crl.CallRail,
        account_id: str,
        json_data: typing.Dict[str, typing.Any]
    ) -> Tag:
        """
        Deserialize JSON data to a Tag object.
        """
        json_data['created_at'] = dateparser.parse(json_data['created_at'])

        return cls(
            api_client=api_client,
            account_id=account_id,
            id=json_data['id'],
            name=json_data['name'],
            tag_level=json_data['tag_level'],
            color=json_data['color'],
            background_color=json_data['background_color'],
            company_id=json_data['company_id'],
            status=json_data['status'],
            created_at=json_data['created_at']
        )
    
    def update(
            self,
            name: typing.Optional[str] = None,
            color: typing.Optional[str] = None,
            disabled: typing.Optional[typing.Union[str, bool]] = None
    ) -> None:
        """
        Updates the Tag.
        """
        if name:
            self.name = name
        if color:
            self.color = color
        if disabled == 'true' or disabled == True:
            self.disabled = True

        body: typing.Dict[str, typing.Any] = {
            'name': name,
            'color': color,
            'disabled': disabled
        }

        response: typing.Dict[str, typing.Any] = typing.cast(
            typing.Dict[str, typing.Any],
            self.api_client._put(
                endpoint=f'a/{self.account_id}',
                path=f'tags/{self.id}.json',
                data=body
            )
        )

        for key, value in response.items():
            if key not in self.__class__.__annotations__.keys():
                raise AttributeError(f'{key} is not a valid attribute for {self.__class__.__name__}')
            else:
                setattr(self, key, value)
    
    def delete(self) -> None:
        """
        Deletes the Tag.
        """
        self.api_client._delete(
            endpoint=f'a/{self.account_id}',
            path=f'tags/{self.id}.json'
        )

    


from __future__ import annotations

import datetime as dt
from dateutil import parser as dateparser
import pycallrail.base as base
import pycallrail.callrail as crl
import typing
import typing_extensions
import requests


class Call(base.CallRailBase):
    """
    Represents a CallRail call.
    
    More information: https://apidocs.callrail.com/#calls
    """

    # These fields are always returned by the API
    answered: bool
    business_phone_number: typing.Optional[str]
    customer_city: str
    customer_country: str
    customer_name: str
    customer_phone_number: str
    customer_state: str
    direction: str
    duration: int
    id: str
    recording: typing.Optional[str]
    recording_duration: typing.Optional[str]
    recording_player: typing.Optional[str]
    start_time: dt.datetime
    tracking_phone_number: str
    voicemail: bool
    
    # User Requested Fields
    call_type: typing.Optional[str]
    company_id: typing.Optional[str]
    company_name: typing.Optional[str]
    company_time_zone: typing.Optional[str]
    created_at: typing.Optional[str]
    device_type: typing.Optional[str]
    first_call: typing.Optional[bool]
    formatted_call_type: typing.Optional[str]
    formatted_customer_location: typing.Optional[str]
    formatted_business_phone_number: typing.Optional[str]
    formatted_customer_name: typing.Optional[str]
    prior_calls: typing.Optional[int]
    formatted_customer_name_or_phone_number: typing.Optional[str]
    formatted_customer_phone_number: typing.Optional[str]
    formatted_duration: typing.Optional[str]
    formatted_tracking_phone_number: typing.Optional[str]
    formatted_tracking_source: typing.Optional[str]
    formatted_value: typing.Optional[str]
    good_lead_call_id: typing.Optional[int]
    good_lead_call_time: typing.Optional[str]
    lead_status: typing.Optional[str]
    note: typing.Optional[str]
    source: typing.Optional[str]
    source_name: typing.Optional[str]
    tags: typing.Optional[typing.List[typing.Any]]
    total_calls: typing.Optional[int]
    value: typing.Optional[str]
    waveforms: typing.Optional[typing.List[typing.Any]]
    tracker_id: typing.Optional[str]
    speaker_percent: typing.Optional[typing.Dict[str, typing.Any]]
    keywords: typing.Optional[str]
    medium: typing.Optional[str]
    campaign: typing.Optional[str]
    referring_url: typing.Optional[str]
    landing_page_url: typing.Optional[str]
    last_requested_url: typing.Optional[str]
    referrer_domain: typing.Optional[str]
    utm_source: typing.Optional[str]
    utm_medium: typing.Optional[str]
    utm_term: typing.Optional[str]
    utm_content: typing.Optional[str]
    utm_campaign: typing.Optional[str]
    utma: typing.Optional[str]
    utmb: typing.Optional[str]
    utmc: typing.Optional[str]
    utmv: typing.Optional[str]
    utmz: typing.Optional[str]
    ga: typing.Optional[str]
    gclid: typing.Optional[str]
    fbclid: typing.Optional[str]
    msclkid: typing.Optional[str]
    milestones: typing.Optional[typing.Any]
    timeline_url: typing.Optional[str]
    keywords_spotted: typing.Optional[typing.List[typing.Any]]
    call_highlights: typing.Optional[typing.List[typing.Any]]
    agent_email: typing.Optional[str]
    keypad_entries: typing.Optional[typing.MutableMapping[str, typing.Any]]


    def __init__(
            self,
            api_client: crl.CallRail,
            account_id: str,
            **kwargs
    ) -> None:
        super(Call, self).__init__()
        self.api_client: crl.CallRail = api_client
        self.account_id: str = account_id

        for key, value in kwargs.items():
            if key not in self.__class__.__annotations__.keys():
                raise AttributeError(f'{key} is not a valid attribute for {self.__class__.__name__}.')
            else:
                setattr(self, key, value)
    
    @classmethod
    def from_json(
        cls,
        api_client: crl.CallRail,
        account_id: str,
        json_data: typing.Dict[str, typing.Any]
    ) -> Call:
        """
        Deserialize JSON data to a Call object.
        
        :param api_client: The CallRail API client.
        :param json_data: The JSON data to deserialize.
        """

        # first lets convert the start_time to a datetime
        json_data['start_time'] = dateparser.parse(json_data['start_time'])

        return cls(api_client, account_id, **json_data)
    
    
    def update(
            self, 
            tags: typing.Optional[typing.List[typing.Any]] = None,
            note: typing.Optional[str] = None,
            value: typing.Optional[str] = None,
            lead_status: typing.Optional[str] = None,
            append_tags: typing.Optional[bool] = None,
            customer_name: typing.Optional[str] = None,
            spam: typing.Optional[bool] = None
        ) -> None:
        """
        Update a Call object.
        More information: https://apidocs.callrail.com/#updating-a-call
        """
        # update the attributes
        if append_tags and tags and hasattr(self, 'tags'):
            self.tags.extend(tags) # type: ignore
        else:
            self.tags = tags
        if note:
            self.note = note
        if value:
            self.value = value
        if lead_status:
            self.lead_status = lead_status
        if customer_name:
            self.customer_name = customer_name
        if spam:
            self.spam = spam

        # form the request body
        body: typing.Dict[str, typing.Any] = {
            'tags': tags,
            'note': note,
            'value': value,
            'lead_status': lead_status,
            'append_tags': append_tags,
            'customer_name': customer_name,
            'spam': spam
        }

        # send the request
        response_data: typing.Dict[str, typing.Any] = typing.cast(
            typing.Dict[str, typing.Any], 
            self.api_client._put(
                endpoint = f'a/{self.account_id}',
                path=f'calls/{self.id}.json',
                data=body
            )
        )

        # update the attributes
        for key, value in response_data.items():
            # raise an error if the attribute doesn't exist
            if key not in self.__class__.__annotations__.keys():
                raise AttributeError(f'{key} is not a valid attribute for {self.__class__.__name__}.')
            else:
                setattr(self, key, value)
        
        
        

    def get_recording(self) -> typing.Union[bytes, None]:
        """
        Get the recording of the call.
        More information: https://apidocs.callrail.com/#get-the-recording-of-the-call

        If no recording exists, None is returned.
        """
        if self.recording:
            with self.api_client.session.get(url=self.recording) as response:
                return response.content
        else:
            return None


import pytest
import pytest_mock
import requests_mock
import requests

from pycallrail.callrail import CallRail
from pycallrail.objects.accounts import Account
import typing




# Tests that list_accounts method returns a list of Account objects. 
def test_list_accounts_happy_path(requests_mock: requests_mock.Mocker) -> None:
    # Setup
    api_key = 'test_api_key'
    api_client = CallRail(api_key=api_key)
    mock_response: typing.Dict[str, typing.Any] = {
        "page": 1,
        "per_page": 100,
        "total_pages": 1,
        "total_records": 2,
        "accounts": [
            {
                "id": "ACC8154748ae6bd4e278a7cddd38a662f4f",
                "name": "Last Mile Metrics",
                "outbound_recording_enabled": True,
                "hipaa_account": False
            },
            {
                "id": "ACC8154748ae6bd4e278a7cddd38a662d4d",
                "name": "CallRail",
                "outbound_recording_enabled": True,
                "hipaa_account": False
            }
        ]   
    }
    
    requests_mock.get('https://api.callrail.com/v3/a.json', json=mock_response)

    # Execution
    accounts: typing.List[Account] = api_client.list_accounts()

    # Assertion
    assert len(accounts) == 2
    assert accounts[0].id == 'ACC8154748ae6bd4e278a7cddd38a662f4f'
    assert accounts[1].id == 'ACC8154748ae6bd4e278a7cddd38a662d4d'
    assert accounts[0].name == 'Last Mile Metrics'
    assert accounts[1].name == 'CallRail'
    assert accounts[0].outbound_recording_enabled == True
    assert accounts[1].outbound_recording_enabled == True
    assert accounts[0].hipaa_account == False
    assert accounts[1].hipaa_account == False

# Tests that get_account returns an Account object. 
def test_get_account_happy_path(requests_mock: requests_mock.Mocker) -> None:
    # Setup
    api_key = 'test_api_key'
    api_client = CallRail(api_key=api_key)
    mock_response: typing.Dict[str, typing.Any] = {
        "id": "ACC8154748ae6bd4e278a7cddd38a662f4f",
        "name": "Last Mile Metrics",
        "outbound_recording_enabled": True,
        "hipaa_account": False
    }
    
    requests_mock.get('https://api.callrail.com/v3/a/ACC8154748ae6bd4e278a7cddd38a662f4f.json', json=mock_response)

    # Execution
    account: Account = api_client.get_account(account_id='ACC8154748ae6bd4e278a7cddd38a662f4f')

    # Assertion
    assert account.id == 'ACC8154748ae6bd4e278a7cddd38a662f4f'
    assert account.name == 'Last Mile Metrics'
    assert account.outbound_recording_enabled == True
    assert account.hipaa_account == False

# Tests that _post handles missing response_data_key. 
def test__post_edge_case(requests_mock: requests_mock.Mocker) -> None:
    # Setup
    api_key = 'test_api_key'
    cr = CallRail(api_key=api_key)
    data_post: typing.Dict[str, typing.Any] = {
        "caller_id": "+17703334455",
        "business_phone_number": "+14045556666",
        "customer_phone_number": "+14044442233",
        "recording_enabled": True,
        "outbound_greeting_recording_url": "http://www.test.com/greeting.mp3",
        "outbound_greeting_text": "These are not the droids you are looking for."
    }

    mock_response: typing.Dict[str, typing.Any] = {
        "answered": None,
        "business_phone_number": "+19044567899",
        "customer_city": "Atlanta",
        "customer_country": "US",
        "customer_name": None,
        "customer_phone_number": "+14703444700",
        "customer_state": "GA",
        "direction": "outbound",
        "duration": None,
        "id": "CAL8154748ae6bd4e278a7cddd38a662f4f",
        "recording": None,
        "recording_duration": None,
        "start_time": "2017-02-22T15:02:24.916-05:00",
        "tracking_phone_number": "+19044567899",
        "voicemail": False
    }
    
    requests_mock.post('https://api.callrail.com/v3/a/12345/calls.json', json=mock_response)

    # Execution
    response: typing.Dict[str, typing.Any] = typing.cast(
        typing.Dict[str, typing.Any], cr._post(endpoint='a/12345', path='calls.json', data=data_post))

    # Assertion
    assert response == mock_response

# Tests that _relative_paginator handles pagination correctly. 
def test__relative_paginator_edge_case(requests_mock: requests_mock.Mocker) -> None:
    # Setup
    api_key = 'test_api_key'
    cr = CallRail(api_key=api_key)
    mock_response_1: typing.Dict[str, typing.Any] = {
        'accounts': [
            {
                'id': '1',
                'name': 'Test Account 1',
                'outbound_recording_enabled': True,
                'hipaa_account': False
            }
        ],
        'has_next_page': True,
        'next_page': 'https://api.callrail.com/v3/a.json?page=2'
    }
    mock_response_2: typing.Dict[str, typing.Any] = {
        'accounts': [
            {
                'id': '2',
                'name': 'Test Account 2',
                'outbound_recording_enabled': False,
                'hipaa_account': True
            }
        ],
        'has_next_page': False
    }
    requests_mock.get('https://api.callrail.com/v3/a.json', json=mock_response_1)
    requests_mock.get('https://api.callrail.com/v3/a.json?page=2', json=mock_response_2)

    # Execution
    response: typing.List[typing.Dict[str, typing.Any]] = cr._relative_paginator(
        response=requests.get('https://api.callrail.com/v3/a.json'), response_data_key='accounts')

    # Assertion
    assert len(response) == 2
    assert response[0]['id'] == '1'
    assert response[1]['id'] == '2'

# Tests handling of invalid API responses. 
def test_list_accounts_general_behavior(requests_mock: requests_mock.Mocker):
    # Setup
    api_key = 'test_api_key'
    cr = CallRail(api_key=api_key)
    mock_response: typing.Dict[str, str] = {
        'error': 'Invalid API key'
    }
    requests_mock.get('https://api.callrail.com/v3/a.json', json=mock_response, status_code=401)

    # Execution and Assertion
    with pytest.raises(requests.exceptions.HTTPError):
        cr.list_accounts()

    # Tests that _get correctly handles query string parameters. 
def test__get_happy_path(requests_mock: requests_mock.Mocker):
    # Setup
    api_key = 'test_api_key'
    cr = CallRail(api_key=api_key)
    mock_response: typing.Dict[str, typing.Any] = {
        "page": 1,
        "per_page": 100,
        "total_pages": 1,
        "total_records": 2,
        "accounts": [
            {
                "id": "ACC8154748ae6bd4e278a7cddd38a662f4f",
                "name": "Last Mile Metrics",
                "outbound_recording_enabled": True,
                "hipaa_account": False
            },
            {
                "id": "ACC8154748ae6bd4e278a7cddd38a662d4d",
                "name": "CallRail",
                "outbound_recording_enabled": True,
                "hipaa_account": False
            }
        ]   
    }

    requests_mock.get('https://api.callrail.com/v3/a.json?sorting=name', json=mock_response)

    # Execution
    response = cr._get(endpoint='a.json', response_data_key='accounts', params={'sorting': 'name'})

    # Assertion
    assert response == mock_response['accounts']

# Tests that _post handles missing response_data_key.  
def test__post_happy_path(requests_mock: requests_mock.Mocker) -> None:
    # Setup
    api_key = 'test_api_key'
    cr = CallRail(api_key=api_key)
    mock_response = {
        'data': {
            'id': 1,
            'name': 'Test Account',
            'outbound_recording_enabled': True,
            'hipaa_account': False
        }
    }

    requests_mock.post('https://api.callrail.com/v3/test_endpoint', json=mock_response)

    # Test
    response = cr._post(endpoint='test_endpoint', response_data_key='data')

    # Assert
    assert response == mock_response['data']
    assert requests_mock.last_request.headers['Authorization'] == f'Token token="{api_key}"'

# Tests handling of API errors and exceptions.  
def test_get_account_general_behavior(requests_mock: requests_mock.Mocker) -> None:
    # Test handling of API errors and exceptions in get_account method
    account_id = '123'
    url = f'https://api.callrail.com/v3/a/{account_id}.json'
    requests_mock.get(url, status_code=404)

    callrail = CallRail(api_key='test')
    with pytest.raises(requests.exceptions.HTTPError):
        callrail.get_account(account_id=account_id)

# Tests handling of unexpected input parameters in CallRail constructor.  
def test_CallRail_constructor_general_behavior() -> None:
    # Test handling of unexpected input parameters in CallRail constructor
    with pytest.raises(TypeError):
        CallRail(api_key='test', invalid_param=True)


import pytest
import pytest_mock
import requests_mock
import requests

from pycallrail.callrail import CallRail
from pycallrail.objects.companies import Company
import typing
import logging
import datetime as dt

# Tests creating a Company object with valid parameters. 
def test_create_company_valid_params(mocker: pytest_mock.MockerFixture) -> None:
    # Arrange
    api_client = mocker.Mock(spec=CallRail)
    account_id = '123'
    company_data = {
        "id": "COM8154748ae6bd4e278a7cddd38a662f4f",
        "name": "Widget Shop",
        "status": "active",
        "time_zone": "America/New_York",
        "created_at": "2017-01-26T22:58:04.184Z",
        "disabled_at": None,
        "dni_active": None,
        "script_url": "//cdn.callrail.com/companies/279054151/a74c824140d67442debd/12/swap.js",
        "callscribe_enabled": False,
        "lead_scoring_enabled": False,
        "swap_exclude_jquery": None,
        "swap_ppc_override": None,
        "swap_landing_override": None,
        "swap_cookie_duration": 365,
        "callscore_enabled": False,
        "keyword_spotting_enabled": False
    }

    # Act
    company = Company(api_client, account_id, **company_data)

    # Assert
    assert company.id == company_data['id']
    assert company.name == company_data['name']
    assert company.status == company_data['status']
    assert company.time_zone == company_data['time_zone']
    assert company.created_at == company_data['created_at']
    assert company.disabled_at == company_data['disabled_at']
    assert company.dni_active == company_data['dni_active']
    assert company.script_url == company_data['script_url']
    assert company.callscore_enabled == company_data['callscore_enabled']
    assert company.lead_scoring_enabled == company_data['lead_scoring_enabled']
    assert company.swap_exclude_jquery == company_data['swap_exclude_jquery']
    assert company.swap_ppc_override == company_data['swap_ppc_override']
    assert company.swap_landing_override == company_data['swap_landing_override']
    assert company.swap_cookie_duration == company_data['swap_cookie_duration']
    assert company.callscribe_enabled == company_data['callscribe_enabled']
    assert company.keyword_spotting_enabled == company_data['keyword_spotting_enabled']

# Tests updating a Company object with valid parameters. 
def test_update_company_valid_params(mocker: pytest_mock.MockerFixture) -> None:
    # Arrange
    api_client = mocker.Mock(spec=CallRail)
    account_id = '123'
    company_data = {
        "id": "COM8154748ae6bd4e278a7cddd38a662f4f",
        "name": "Widget Shop",
        "status": "active",
        "time_zone": "America/New_York",
        "created_at": "2017-01-26T22:58:04.184Z",
        "disabled_at": None,
        "dni_active": None,
        "script_url": "//cdn.callrail.com/companies/279054151/a74c824140d67442debd/12/swap.js",
        "callscribe_enabled": False,
        "lead_scoring_enabled": False,
        "swap_exclude_jquery": None,
        "swap_ppc_override": None,
        "swap_landing_override": None,
        "swap_cookie_duration": 365,
        "callscore_enabled": False,
        "keyword_spotting_enabled": False
    }
    company = Company(api_client, account_id, **company_data)

    updated_name = 'Updated Company Name'

    # Act
    company.update(name=updated_name)

    # Assert
    assert company.name == updated_name

# Tests creating a Company object with missing parameters. 
def test_create_company_missing_params(mocker: pytest_mock.MockerFixture) -> None:
    # Arrange
    api_client = mocker.Mock(spec=CallRail)
    account_id = '123'
    # Act and Assert
    with pytest.raises(TypeError):
        Company(api_client, account_id)

# Tests creating a Company object with invalid parameters. 
def test_create_company_invalid_params(mocker: pytest_mock.MockerFixture) -> None:
    # Arrange
    api_client = mocker.Mock()
    account_id = '123'

    # Act and Assert
    with pytest.raises(AttributeError):
        Company(api_client, account_id, invalid_param='invalid')

# Tests deleting a Company object successfully. 
def test_delete_company(mocker: pytest_mock.MockerFixture):
    # Arrange
    api_client = mocker.Mock(spec=CallRail)
    account_id = '123'
    company_data = {
        'id': '456',
        'name': 'Test Company',
        'status': 'active',
        'time_zone': 'America/New_York',
        'created_at': dt.datetime.now(),
        'disabled_at': None,
        'dni_active': False,
        'script_url': 'https://example.com/script.js',
        'callscore_enabled': True,
        'lead_scoring_enabled': False,
        'swap_exclude_jquery': None,
        'swap_ppc_override': None,
        'swap_landing_override': None,
        'swap_cookie_duration': None,
        'callscribe_enabled': True,
        'keyword_spotting_enabled': None,
        'form_capture': True
    }
    company = Company(api_client, account_id, **company_data)

    # Act
    company.delete()

    # Assert
    api_client._delete.assert_called_once_with(endpoint=f'/a/{account_id}', path=f'/companies/{company.id}.json')


from urllib.parse import urljoin
import typing
def build_url(base_url: str, endpoint: str, path: typing.Optional[str] = None ) -> str:
    
    url_result: str = urljoin(base_url, endpoint)

    if path:
        # Ensure the endpoint has a trailing slash
        if not url_result.endswith('/'):
            url_result += '/'

        # Ensure the path doesn't have a leading slash
        if path.startswith('/'):
            path = path[1:]

        url_result: str = urljoin(url_result, path) # type: ignore

    return url_result


import pytest
import pytest_mock
import requests_mock
import requests

from pycallrail.callrail import CallRail
from pycallrail.objects.calls import Call
import typing
import logging
import datetime as dt

# Tests that a Call object can be created with valid arguments. 
def test_create_call_with_valid_arguments(mocker: pytest_mock.MockerFixture) -> None:
    # Arrange
    api_client = mocker.Mock(spec=CallRail)
    account_id = '123'
    call_data: typing.Dict[str, typing.Any] =  {
        "answered": False,
        "business_phone_number": None,
        "customer_city": "New York City",
        "customer_country": "US",
        "customer_name": "Jimmy Pesto, Sr.",
        "customer_phone_number": "+13036231131",
        "customer_state": "NY",
        "direction": "inbound",
        "duration": 4,
        "id": "CAL8154748ae6bd4e278a7cddd38a662f4f",
        "recording": "https://api.callrail.com/v3/a/227799611/calls/111222333/recording.json",
        "recording_duration": "27",
        "recording_player": "https://app.callrail.com/calls/111222333/recording?access_key=3b91eb7f7cc08a4d01ed",
        "start_time": "2017-01-24T11:27:48.119-05:00",
        "tracking_phone_number": "+13038163491",
        "voicemail": False,
        "agent_email": "gil@televised.com"
    }

    # Act
    call = Call(api_client, account_id, **call_data)

    # Assert
    assert call.answered == call_data['answered']
    assert call.customer_city == call_data['customer_city']
    assert call.customer_country == call_data['customer_country']
    assert call.customer_name == call_data['customer_name']
    assert call.customer_phone_number == call_data['customer_phone_number']
    assert call.customer_state == call_data['customer_state']
    assert call.direction == call_data['direction']
    assert call.duration == call_data['duration']
    assert call.id == call_data['id']
    assert call.start_time == call_data['start_time']
    assert call.tracking_phone_number == call_data['tracking_phone_number']
    assert call.voicemail == call_data['voicemail']

# Tests that a Call object can be created with valid arguments. 
def test_call_creation_two_instances(mocker: pytest_mock.MockerFixture) -> None:
    # Arrange
    api_client = mocker.Mock(spec=CallRail)
    account_id = '123'
    call_data: typing.Dict[str, typing.Any] =  {
        "answered": False,
        "business_phone_number": None,
        "customer_city": "New York City",
        "customer_country": "US",
        "customer_name": "Jimmy Pesto, Sr.",
        "customer_phone_number": "+13036231131",
        "customer_state": "NY",
        "direction": "inbound",
        "duration": 4,
        "id": "CAL8154748ae6bd4e278a7cddd38a662f4f",
        "recording": "https://api.callrail.com/v3/a/227799611/calls/111222333/recording.json",
        "recording_duration": "27",
        "recording_player": "https://app.callrail.com/calls/111222333/recording?access_key=3b91eb7f7cc08a4d01ed",
        "start_time": "2017-01-24T11:27:48.119-05:00",
        "tracking_phone_number": "+13038163491",
        "voicemail": False,
        "agent_email": "gil@televised.com"
    }

    # Act
    call = Call(api_client, account_id, **call_data)

    call_data_2: typing.Dict[str, typing.Any] =  {
        "answered": True,
        "business_phone_number": None,
        "customer_city": "Toronto",
        "customer_country": "CA",
        "customer_name": "Jimmy Pesto, Sr.",
        "customer_phone_number": "+13036231131",
        "customer_state": "NY",
        "direction": "inbound",
        "duration": 4,
        "id": "CAL8154748ae6bd4e278a7cddd38a662f4f",
        "recording": "https://api.callrail.com/v3/a/227799611/calls/111222333/recording.json",
        "recording_duration": "27",
        "recording_player": "https://app.callrail.com/calls/111222333/recording?access_key=3b91eb7f7cc08a4d01ed",
        "start_time": "2017-01-24T11:27:48.119-05:00",
        "tracking_phone_number": "+13038163491",
        "voicemail": False,
        "agent_email": "gil@televised.com"
    }

    call_2 = Call(api_client, account_id, **call_data_2)

    # Assert
    assert call.answered == call_data['answered']
    assert call.customer_city == call_data['customer_city']
    assert call.customer_country == call_data['customer_country']
    assert call.customer_name == call_data['customer_name']
    assert call.customer_phone_number == call_data['customer_phone_number']
    assert call.customer_state == call_data['customer_state']
    assert call.direction == call_data['direction']
    assert call.duration == call_data['duration']
    assert call.id == call_data['id']
    assert call.start_time == call_data['start_time']
    assert call.tracking_phone_number == call_data['tracking_phone_number']
    assert call.voicemail == call_data['voicemail']

    assert call_2.answered == call_data_2['answered']
    assert call_2.customer_city == call_data_2['customer_city']
    assert call_2.customer_country == call_data_2['customer_country']

# Tests that a Call object can be updated with valid arguments. 
def test_update_call_with_valid_arguments(requests_mock: requests_mock.Mocker) -> None:
    # Arrange
    api_client = CallRail(api_key='123')
    account_id = '123'
    
    call_data =  {
        "answered": False,
        "business_phone_number": None,
        "customer_city": "Denver",
        "customer_country": "US",
        "customer_name": "James Smith",
        "customer_phone_number": "+13036231131",
        "customer_state": "CO",
        "direction": "inbound",
        "duration": 4,
        "id": "CAL8154748ae6bd4e278a7cddd38a662f4f",
        "recording": "https://api.callrail.com/v3/a/227799611/calls/213472384/recording.json",
        "recording_duration": "27",
        "start_time": "2017-01-24T11:27:48.119-05:00",
        "tracking_phone_number": "+13038163491",
        "voicemail": False
    }

    call = Call(api_client, account_id, **call_data)

    update_data = {
        "note": "Call customer back tomorrow",
        "tags": ["New Client"],
        "lead_status": "good_lead",
        "value": "$1.00",
        "append_tags": True,
        "customer_name": "James Smith"
    }

    # Act
    requests_mock.put(
        url=f'https://api.callrail.com/v3/a/{account_id}/calls/{call.id}.json',
        json= call_data 
    )


    call.update(
        tags = update_data['tags'], 
        note = update_data['note'], 
        value = update_data['value'], 
        append_tags = update_data['append_tags'], 
        customer_name = update_data['customer_name'])

    # Assert
    assert call.tags == update_data['tags']
    assert call.note == update_data['note']
    assert call.value == update_data['value']
    assert call.customer_name == update_data['customer_name']

# Tests that a Call object cannot be created with missing or invalid required arguments. 
def test_create_call_with_invalid_arguments(mocker: pytest_mock.MockerFixture) -> None:
    # Arrange
    api_client = mocker.Mock(spec=CallRail)
    account_id = '123'
    call_data = {
        "answered": False,
        "business_phone_number": None,
        "customer_city": "Denver",
        "customer_country": "US",
        "customer_name": "James Smith",
        "customer_phone_number": "+13036231131",
        "customer_state": "CO",
        "direction": "inbound",
        "duration": 4,
        "recording": "https://api.callrail.com/v3/a/227799611/calls/213472384/recording.json",
        "recording_duration": "27",
        "start_time": "2017-01-24T11:27:48.119-05:00",
        "tracking_phone_number": "+13038163491",
        "invalid_argument": True
    } 

    # Act & Assert
    with pytest.raises(AttributeError):
        Call(api_client, account_id, **call_data)

def test_update_call_with_invalid_arguments(mocker: pytest_mock.MockerFixture) -> None:
    # Arrange
    api_client = mocker.Mock(spec=CallRail)
    account_id = '123'
    call_data = {
        "answered": False,
        "business_phone_number": None,
        "customer_city": "Denver",
        "customer_country": "US",
        "customer_name": "James Smith",
        "customer_phone_number": "+13036231131",
        "customer_state": "CO",
        "direction": "inbound",
        "duration": 4,
        "recording": "https://api.callrail.com/v3/a/227799611/calls/213472384/recording.json",
        "recording_duration": "27",
        "start_time": "2017-01-24T11:27:48.119-05:00",
        "tracking_phone_number": "+13038163491"
    }

    call = Call(api_client, account_id, **call_data)

    with pytest.raises(Exception):
        call.update(invalid_argument = True)

# Tests that the recording of a call can be retrieved with a valid recording URL. 
def test_get_recording_with_valid_recording_url(requests_mock: requests_mock.Mocker) -> None:
    # Arrange
    api_client = CallRail(api_key='123')
    account_id = '123'
    call_data = {
        'answered': True,
        'customer_city': 'New York',
        'customer_country': 'USA',
        'customer_name': 'John Doe',
        'customer_phone_number': '123-456-7890',
        'customer_state': 'NY',
        'direction': 'inbound',
        'duration': 60,
        'id': '456',
        'start_time': '2017-01-24T11:27:48.119-05:00',
        'tracking_phone_number': '987-654-3210',
        'voicemail': False,
        'recording': 'http://example.com/recording.mp3'
    }
    call = Call(api_client, account_id, **call_data)
    recording_content = b'This is a recording'

    # Mock the response from the recording URL
    requests_mock.get(
        url='http://example.com/recording.mp3',
        content=recording_content
    )

    # Act
    recording = call.get_recording()

    # Assert
    assert recording == recording_content

# Tests that None is returned when attempting to retrieve the recording of a call with no recording URL. 
def test_get_recording_with_no_recording_url(mocker: pytest_mock.MockerFixture) -> None:
    # Arrange
    api_client = mocker.Mock(spec=CallRail)
    account_id = '123'
    call_data = {
        'answered': True,
        'customer_city': 'New York',
        'customer_country': 'USA',
        'customer_name': 'John Doe',
        'customer_phone_number': '123-456-7890',
        'customer_state': 'NY',
        'direction': 'inbound',
        'duration': 60,
        'id': '456',
        'start_time': '2017-01-24T11:27:48.119-05:00',
        'tracking_phone_number': '987-654-3210',
        'voicemail': False,
        'recording': None
    }
    call = Call(api_client, account_id, **call_data)

    # Act
    recording = call.get_recording()

    # Assert
    assert recording is None

# Tests that JSON data can be successfully deserialized to a Call object.  
def test_deserialize_json_data_to_call_object(mocker: pytest_mock.MockerFixture) -> None:
    # Setup
    api_client = mocker.Mock(spec=CallRail)
    account_id = '123'
    json_data = {
        'answered': True,
        'business_phone_number': '555-555-5555',
        'customer_city': 'New York',
        'customer_country': 'USA',
        'customer_name': 'John Doe',
        'customer_phone_number': '555-555-5555',
        'customer_state': 'NY',
        'direction': 'inbound',
        'duration': 60,
        'id': '456',
        'recording': 'http://example.com/recording.mp3',
        'recording_duration': '00:01:00',
        'recording_player': 'http://example.com/recording_player.mp3',
        'start_time': '2022-01-01T00:00:00Z',
        'tracking_phone_number': '555-555-5555',
        'voicemail': False,
        'call_type': 'sales',
        'company_id': '789',
        'company_name': 'Acme Inc.',
        'company_time_zone': 'America/New_York',
        'created_at': '2022-01-01T00:00:00Z',
        'device_type': 'mobile',
        'first_call': True,
        'formatted_call_type': 'Sales Call',
        'formatted_customer_location': 'New York, NY, USA',
        'formatted_business_phone_number': '(555) 555-5555',
        'formatted_customer_name': 'John Doe',
        'prior_calls': 2,
        'formatted_customer_name_or_phone_number': '(555) 555-5555',
        'formatted_customer_phone_number': '(555) 555-5555',
        'formatted_duration': '1m 0s',
        'formatted_tracking_phone_number': '(555) 555-5555',
        'formatted_tracking_source': 'Google Ads',
        'formatted_value': '$100',
        'good_lead_call_id': 123,
        'good_lead_call_time': '2022-01-01T00:00:00Z',
        'lead_status': 'qualified',
        'note': 'This was a good call.',
        'source': 'google',
        'source_name': 'Google Ads',
        'tags': ['tag1', 'tag2'],
        'total_calls': 10,
        'value': 100,
        'waveforms': [{'time': 0, 'value': 0}, {'time': 1, 'value': 1}],
        'tracker_id': 'abc123',
        'speaker_percent': {'speaker1': 50, 'speaker2': 50},
        'keywords': 'important keywords',
        'medium': 'cpc',
        'campaign': 'campaign1',
        'referring_url': 'http://example.com/referrer',
        'landing_page_url': 'http://example.com/landing',
        'last_requested_url': 'http://example.com/requested',
        'referrer_domain': 'example.com',
        'utm_source': 'google',
        'utm_medium': 'cpc',
        'utm_term': 'term1',
        'utm_content': 'content1',
        'utm_campaign': 'campaign1',
        'utma': '123456789.1234567890.1234567890.1234567890.1',
        'utmb': '123456789.1.10.1234567890',
        'utmc': '',
        'utmv': '',
        'utmz': '',
        'ga': '',
        'gclid': '',
        'fbclid': '',
        'msclkid': '',
        'milestones': {'milestone1': '2022-01-01T00:00:00Z', 'milestone2': '2022-01-01T00:01:00Z'},
        'timeline_url': 'http://example.com/timeline',
        'keywords_spotted': ['keyword1', 'keyword2'],
        'call_highlights': [{'start_time': '2022-01-01T00:00:00Z', 'end_time': '2022-01-01T00:01:00Z', 'text': 'highlighted text'}],
        'agent_email': 'johndoe@example.com',
        'keypad_entries': {'1': 2, '2': 3}
    }

    # Exercise
    call = Call.from_json(api_client, account_id, json_data)

    # Verify
    assert isinstance(call, Call)
    assert call.answered == True
    assert call.business_phone_number == '555-555-5555'
    assert call.customer_city == 'New York'
    assert call.customer_country == 'USA'
    assert call.customer_name == 'John Doe'
    assert call.customer_phone_number == '555-555-5555'
    assert call.customer_state == 'NY'
    assert call.direction == 'inbound'
    assert call.duration == 60
    assert call.id == '456'
    assert call.recording == 'http://example.com/recording.mp3'
    assert call.recording_duration == '00:01:00'
    assert call.recording_player == 'http://example.com/recording_player.mp3'
    assert isinstance(call.start_time, dt.datetime)
    assert call.start_time.isoformat() == '2022-01-01T00:00:00+00:00'
    assert call.tracking_phone_number == '555-555-5555'
    assert call.voicemail == False
    assert call.call_type == 'sales'
    assert call.company_id == '789'
    assert call.company_name == 'Acme Inc.'
    assert call.company_time_zone == 'America/New_York'
    assert call.created_at == '2022-01-01T00:00:00Z'
    assert call.device_type == 'mobile'
    assert call.first_call == True
    assert call.formatted_call_type == 'Sales Call'
    assert call.formatted_customer_location == 'New York, NY, USA'
    assert call.formatted_business_phone_number == '(555) 555-5555'
    assert call.formatted_customer_name == 'John Doe'
    assert call.prior_calls == 2
    assert call.formatted_customer_name_or_phone_number == '(555) 555-5555'
    assert call.formatted_customer_phone_number == '(555) 555-5555'
    assert call.formatted_duration == '1m 0s'
    assert call.formatted_tracking_phone_number == '(555) 555-5555'
    assert call.formatted_tracking_source == 'Google Ads'
    assert call.formatted_value == '$100'
    assert call.good_lead_call_id == 123
    assert call.good_lead_call_time == '2022-01-01T00:00:00Z'
    assert call.lead_status == 'qualified'
    assert call.note == 'This was a good call.'
    assert call.source == 'google'
    assert call.source_name == 'Google Ads'
    assert call.tags == ['tag1', 'tag2']
    assert call.total_calls == 10
    assert call.value == 100
    assert isinstance(call.waveforms, list)
    assert call.waveforms == [{'time': 0, 'value': 0}, {'time': 1, 'value': 1}]
    assert call.tracker_id == 'abc123'
    assert call.speaker_percent == {'speaker1': 50, 'speaker2': 50}
    assert call.keywords == 'important keywords'
    assert call.medium == 'cpc'
    assert call.campaign == 'campaign1'
    assert call.referring_url == 'http://example.com/referrer'
    assert call.landing_page_url == 'http://example.com/landing'
    assert call.last_requested_url == 'http://example.com/requested'
    assert call.referrer_domain == 'example.com'
    assert call.utm_source == 'google'
    assert call.utm_medium == 'cpc'
    assert call.utm_term == 'term1'
    assert call.utm_content == 'content1'
    assert call.utm_campaign == 'campaign1'
    assert call.utma == '123456789.1234567890.1234567890.1234567890.1'
    assert call.utmb == '123456789.1.10.1234567890'
    assert call.utmc == ''
    assert call.utmv == ''
    assert call.utmz == ''
    assert call.ga == ''
    assert call.gclid == ''
    assert call.fbclid == ''
    assert call.msclkid == ''
    assert isinstance(call.milestones, dict)
    assert call.milestones == {'milestone1': '2022-01-01T00:00:00Z', 'milestone2': '2022-01-01T00:01:00Z'}
    assert call.timeline_url == 'http://example.com/timeline'
    assert call.keywords_spotted == ['keyword1', 'keyword2']
    assert isinstance(call.call_highlights, list)
    assert call.call_highlights == [{'start_time': '2022-01-01T00:00:00Z', 'end_time': '2022-01-01T00:01:00Z', 'text': 'highlighted text'}]
    assert call.agent_email == 'johndoe@example.com'
    assert isinstance(call.keypad_entries, dict)
    assert call.keypad_entries == {'1': 2, '2': 3}


import pytest
import pytest_mock
import requests_mock
import requests

from pycallrail.callrail import CallRail
from pycallrail.objects.form_submissions import FormSubmission
import typing
import logging
import datetime as dt
import typeguard

# Tests that a FormSubmission object can be created with valid arguments. 
def test_create_form_submission_valid_args(mocker: pytest_mock.MockerFixture) -> None:
    # Arrange
    api_client = mocker.Mock(spec=CallRail)
    account_id = '123'
    form_data = {'name': 'John Doe', 'email': 'johndoe@example.com'}
    form_url = 'https://example.com/form'
    landing_page_url = 'https://example.com/landing'
    referrer = 'https://google.com'
    referring_url = 'https://google.com/search?q=example'
    submitted_at = dt.datetime.now()
    first_form = True
    customer_phone_number = '123-456-7890'
    customer_name = 'John Doe'
    formatted_customer_phone_number = '+11234567890'
    formatted_customer_name = 'John D.'
    source = 'Google'
    keywords = 'example'
    campaign = 'Example Campaign'
    medium = 'Organic'

    # Act
    form_submission = FormSubmission(
        api_client=api_client,
        account_id=account_id,
        id='456',
        company_id='789',
        person_id='012',
        form_data=form_data,
        form_url=form_url,
        landing_page_url=landing_page_url,
        referrer=referrer,
        referring_url=referring_url,
        submitted_at=submitted_at,
        first_form=first_form,
        customer_phone_number=customer_phone_number,
        customer_name=customer_name,
        formatted_customer_phone_number=formatted_customer_phone_number,
        formatted_customer_name=formatted_customer_name,
        source=source,
        keywords=keywords,
        campaign=campaign,
        medium=medium
    )

    # Assert
    assert isinstance(form_submission, FormSubmission)
    assert form_submission.api_client == api_client
    assert form_submission.account_id == account_id
    assert form_submission.id == '456'
    assert form_submission.company_id == '789'
    assert form_submission.person_id == '012'
    assert form_submission.form_data == form_data
    assert form_submission.form_url == form_url
    assert form_submission.landing_page_url == landing_page_url
    assert form_submission.referrer == referrer
    assert form_submission.referring_url == referring_url
    assert form_submission.submitted_at == submitted_at
    assert form_submission.first_form == first_form
    assert form_submission.customer_phone_number == customer_phone_number
    assert form_submission.customer_name == customer_name
    assert form_submission.formatted_customer_phone_number == formatted_customer_phone_number
    assert form_submission.formatted_customer_name == formatted_customer_name
    assert form_submission.source == source
    assert form_submission.keywords == keywords
    assert form_submission.campaign == campaign
    assert form_submission.medium == medium

# Tests that a FormSubmission object can be updated with valid arguments. 
def test_update_form_submission_valid_args(mocker: pytest_mock.MockerFixture) -> None:
    # Arrange
    api_client = mocker.Mock(spec=CallRail)
    account_id = '123'
    form_submission = FormSubmission(
        api_client=api_client,
        account_id=account_id,
        id='456',
        company_id='789',
        person_id='012',
        form_data={'name': 'John Doe', 'email': 'johndoe@example.com'},
        form_url='https://example.com/form',
        landing_page_url='https://example.com/landing',
        referrer='https://google.com',
        referring_url='https://google.com/search?q=example',
        submitted_at=dt.datetime.now(),
        first_form=True,
        customer_phone_number='123-456-7890',
        customer_name='John Doe',
        formatted_customer_phone_number='+11234567890',
        formatted_customer_name='John D.',
        source='Google',
        keywords='example',
        campaign='Example Campaign',
        medium='Organic'
    )

    api_client._put.return_value = {
        'id': '456',
        'company_id': '789',
        'person_id': '012',
        'form_data': {'name': 'John Doe', 'email': 'johndoe@example.com'},
        'form_url': 'https://example.com/form',
        'landing_page_url': 'https://example.com/landing',
        'referrer': 'https://google.com',
        'referring_url': 'https://google.com/search?q=example',
        'submitted_at': '2021-01-01T00:00:00+00:00',
        'first_form': True
    }
    # Act
    form_submission.update(
        tags=['tag1', 'tag2'],
        note='Example note',
        value=100,
        lead_status='New',
        append_tags=True
    )

    # Assert
    assert form_submission.tags == ['tag1', 'tag2']
    assert form_submission.note == 'Example note'
    assert form_submission.value == 100
    assert form_submission.lead_status == 'New'

# Tests that a FormSubmission object cannot be updated with missing or invalid optional arguments. 
def test_update_form_submission_missing_args(mocker: pytest_mock.MockerFixture):
    # Arrange
    api_client = mocker.Mock(spec=CallRail)
    account_id = '123'
    form_submission = FormSubmission(
        api_client=api_client,
        account_id=account_id,
        id='456',
        company_id='789',
        person_id='012',
        form_data={'name': 'John Doe', 'email': 'johndoe@example.com'},
        form_url='https://example.com/form',
        landing_page_url='https://example.com/landing',
        referrer='https://google.com',
        referring_url='https://google.com/search?q=example',
        submitted_at=dt.datetime.now(),
        first_form=True,
        customer_phone_number='123-456-7890',
        customer_name='John Doe',
        formatted_customer_phone_number='+11234567890',
        formatted_customer_name='John D.',
        source='Google',
        keywords='example',
        campaign='Example Campaign',
        medium='Organic'
    )

    # Act and Assert
    with pytest.raises(TypeError):
        form_submission.update(
            tags=['tag1', 'tag2'],
            note='Example note',
            value=100,
            lead_status='New',
            invalid_arg=True
        )

# Tests that a FormSubmission object is not updated if no changes are made. 
def test_update_form_submission_no_changes(mocker: pytest_mock.MockerFixture):
    # Arrange
    api_client = mocker.Mock(spec=CallRail)
    account_id = '123'
    form_submission = FormSubmission(
        api_client=api_client,
        account_id=account_id,
        id='456',
        company_id='789',
        person_id='012',
        form_data={'name': 'John Doe', 'email': 'johndoe@example.com'},
        form_url='https://example.com/form',
        landing_page_url='https://example.com/landing',
        referrer='https://google.com',
        referring_url='https://google.com/search?q=example',
        submitted_at=dt.datetime.now(),
        first_form=True,
        customer_phone_number='123-456-7890',
        customer_name='John Doe',
        formatted_customer_phone_number='+11234567890',
        formatted_customer_name='John D.',
        source='Google',
        keywords='example',
        campaign='Example Campaign',
        medium='Organic'
    )

    api_client._put.return_value = {
        'id': '456',
        'company_id': '789',
        'person_id': '012',
        'form_data': {'name': 'John Doe', 'email': 'johndoe@example.com'},
        'form_url': 'https://example.com/form',
        'landing_page_url': 'https://example.com/landing',
        'referrer': 'https://google.com',
        'referring_url': 'https://google.com/search?q=example',
        'submitted_at': '2021-01-01T00:00:00+00:00',
        'first_form': True
    }

    # Act
    form_submission.update()

    # Assert
    assert hasattr(form_submission, 'tags') is False
    assert hasattr(form_submission, 'note') is False
    assert hasattr(form_submission, 'value') is False
    assert hasattr(form_submission, 'lead_status') is False

    # Tests that a FormSubmission object can be deserialized from JSON. 
def test_deserialize_form_submission_from_json(mocker: pytest_mock.MockerFixture):
    # Arrange
    api_client = mocker.Mock(spec=CallRail)
    account_id = '123'
    json_data = {
        "id": "FOR8154748ae6bd4e278a7cddd38a662f4f",
        "company_id": "COM8154748ae6bd4e278a7cddd38a662f4f",
        "person_id": "anonymous",
        "form_data": {
            "name": "Graham Armstrong",
            "email": "graham@example.com",
            "phone": "(999) 999-9999",
            "contact_method": "call",
            "best_time_to_call": "evening"
        },
        "form_url": "http://www.uptowndental.com/appointment",
        "landing_page_url": "http://www.uptowndental.com/appointment?campaign=back_to_school&keywords=pediatric%20dentist&gclid=12345abgdef",
        "referrer": "www.google.com",
        "referring_url": "www.google.com",
        "submitted_at": "2017-11-17T07:46:28.000Z",
        "first_form": False,
        "customer_phone_number": "+19999999999",
        "customer_name": "Graham Armstrong",
        "formatted_customer_phone_number": "999-999-9999",
        "formatted_customer_name": "Graham Armstrong",
        "source": "Google Ads",
        "keywords": "pediatric dentist",
        "campaign": "back_to_school",
        "medium": "CPC",
        "milestones": {
            "first_touch": {
                "event_date": "2017-10-22T19:41:45.183-04:00",
                "ad_position": None,
                "campaign": "back_to_school",
                "device": "desktop",
                "keywords": "pediatric dentist",
                "landing": "http://www.uptowndental.com/appointment?campaign=back_to_school&keywords=pediatric%20dentist&gclid=12345abgdef",
                "landing_page_url_params": {
                        "campaign": "back_to_school",
                        "keywords": "pediatric dentist",
                        "gclid": "12345abgdef"
                },
                "match_type": None,
                "medium": "cpc",
                "referrer": "https://www.google.com/",
                "referrer_url_params": {},
                "session_browser": "Chrome",
                "url_utm_params": {},
                "source": "Google Ads"
            },
            "lead_created": {
                "event_date": "2017-10-22T19:41:45.183-04:00",
                "ad_position": None,
                "campaign": "back_to_school",
                "device": "desktop",
                "keywords": "pediatric dentist",
                "landing": "http://www.uptowndental.com/appointment?campaign=back_to_school&keywords=pediatric%20dentist&gclid=12345abgdef",
                "landing_page_url_params": {
                        "campaign": "back_to_school",
                        "keywords": "pediatric dentist",
                        "gclid": "12345abgdef"
                },
                "match_type": None,
                "medium": "cpc",
                "referrer": "https://www.google.com/",
                "referrer_url_params": {},
                "session_browser": "Chrome",
                "url_utm_params": {},
                "source": "Google Ads"
            },
            "qualified": {
                "event_date": "2017-10-22T19:41:45.183-04:00",
                "ad_position": None,
                "campaign": "back_to_school",
                "device": "desktop",
                "keywords": "pediatric dentist",
                "landing": "http://www.uptowndental.com/appointment?campaign=back_to_school&keywords=pediatric%20dentist&gclid=12345abgdef",
                "landing_page_url_params": {
                        "campaign": "back_to_school",
                        "keywords": "pediatric dentist",
                        "gclid": "12345abgdef"
                },
                "match_type": None,
                "medium": "cpc",
                "referrer": "https://www.google.com/",
                "referrer_url_params": {},
                "session_browser": "Chrome",
                "url_utm_params": {},
                "source": "Google Ads"
            },
            "last_touch": {
                "event_date": "2017-11-17T07:46:28.000-04:00",
                "ad_position": None,
                "campaign": None,
                "device": "desktop",
                "keywords": None,
                "landing": "https://www.example.com/",
                "landing_page_url_params": {},
                "match_type": None,
                "medium": "organic",
                "referrer": "https://www.google.com/",
                "referrer_url_params": {},
                "session_browser": "Chrome",
                "url_utm_params": {},
                "source": "Google Organic"
            }
        }
    }

    # Act
    form_submission: FormSubmission = FormSubmission.from_json(
        api_client=api_client,
        account_id=account_id,
        json_data=json_data
    )

    # Assert
    assert isinstance(form_submission, FormSubmission)
    assert form_submission.api_client == api_client
    assert form_submission.account_id == account_id
    assert form_submission.id == json_data['id']
    assert form_submission.company_id == json_data['company_id']
    assert form_submission.person_id == json_data['person_id']
    assert form_submission.form_data == json_data['form_data']
    assert form_submission.form_url == json_data['form_url']
    assert form_submission.landing_page_url == json_data['landing_page_url']
    assert form_submission.referrer == json_data['referrer']
    assert form_submission.referring_url == json_data['referring_url']
    assert isinstance(form_submission.submitted_at, dt.datetime)
    assert form_submission.first_form == json_data['first_form']
    assert form_submission.customer_phone_number == json_data['customer_phone_number']
    assert form_submission.customer_name == json_data['customer_name']
    assert form_submission.formatted_customer_phone_number == json_data['formatted_customer_phone_number']
    assert form_submission.formatted_customer_name == json_data['formatted_customer_name']
    assert form_submission.source == json_data['source']
    assert form_submission.keywords == json_data['keywords']
    assert form_submission.campaign == json_data['campaign']
    assert form_submission.medium == json_data['medium']
    assert form_submission.milestones == json_data['milestones']


from __future__ import annotations

import datetime as dt
from dateutil import parser as dateparser
import pycallrail.base as base
import pycallrail.callrail as crl
import pycallrail.objects.calls as calls
import pycallrail.objects.tags as tags
import pycallrail.objects.companies as companies
import pycallrail.objects.form_submissions as forms
import pycallrail.objects.textmessages as messages
import typing
import logging

class Account(base.CallRailBase):
    """
    Represents a CallRail Account. Class should preferably not be instantiated directly.

    More information: https://apidocs.callrail.com/#accounts
    """

    id: str
    name: str
    outbound_recording_enabled: bool
    hipaa_account: bool

    # Optional User Requested Fields
    numeric_id: typing.Optional[int]

    def __init__(
            self,
            api_client: crl.CallRail,
            id: str,
            name: str,
            outbound_recording_enabled: bool,
            hipaa_account: bool,
            numeric_id: typing.Optional[int] = None
    ) -> None:
        super(Account, self).__init__()
        self.api_client: crl.CallRail = api_client
        self.id = id
        self.name = name
        self.outbound_recording_enabled = outbound_recording_enabled
        self.hipaa_account = hipaa_account

        if numeric_id:
            self.numeric_id = numeric_id

    @classmethod
    def from_json(
        cls,
        api_client: crl.CallRail,
        json_data: typing.Dict[str, typing.Any]
    ) -> Account:
        """
        Deserialize JSON data into an Account object.
        
        :api_client: CallRail API object
        :json_data: JSON data
        """

        account = Account(
            api_client=api_client,
            id=json_data['id'],
            name=json_data['name'],
            outbound_recording_enabled=json_data['outbound_recording_enabled'],
            hipaa_account=json_data['hipaa_account']
        )

        if 'numeric_id' in json_data:
            account.numeric_id = json_data['numeric_id']

        return account
    
    #########################
    # Calls
    #########################

    def list_calls(
            self,
            **kwargs
        ) -> typing.Union[typing.List[calls.Call], None]:
        """
        List all calls associated with this account.

        Keyword args acceptable include: Pagination Type, Sorting, Filtering, Field Selection, and Searching.
        Defaults to relative pagination.
        
        More info: https://apidocs.callrail.com/#listing-all-calls
        """

        pagination_type: str = kwargs.get('pagination_type', 'RELATIVE')

        sorting = kwargs.get('sorting', None)
        filtering = kwargs.get('filtering', None)
        searching = kwargs.get('searching', None)
        fields = kwargs.get('fields', None)

        params = {}

        if sorting:
            params['sorting'] = sorting
        if filtering:
            params['filtering'] = filtering
        if searching:
            params['searching'] = searching
        if fields:
            params['fields'] = fields

        if calls_response := typing.cast(
            typing.Union[typing.List[typing.Dict[str, typing.Any]], None],
            self.api_client._get(
                endpoint=f'a/{self.id}',
                response_data_key='calls',
                path='calls.json',
                params=params or None,
                pagination_type=pagination_type,
            ),
        ):
            return [calls.Call.from_json(self.api_client, self.id, call) for call in calls_response]
        else:
            return None

    def get_call(
        self,
        call_id: str,
        fields: typing.Optional[typing.MutableMapping[str, typing.Union[str, typing.Any]]] = None
    ) -> calls.Call:
        """
        Retrieve a single call by ID.
        
        Field selection is supported by this method.
        
        More info: https://apidocs.callrail.com/#retrieving-a-single-call
        """

        if fields:
            data = self.api_client._get(
                endpoint=f'a/{self.id}',
                path=f'calls/{call_id}.json',
                params=fields
            )
        else:
            data = self.api_client._get(
                endpoint=f'a/{self.id}',
                path=f'calls/{call_id}.json'
            )

        return calls.Call.from_json(
            self.api_client, 
            self.id, 
            typing.cast(typing.Dict[str, typing.Any], data)
        )
    
    def create_call(
            self,
            caller_id: int,
            customer_phone_number: str,
            business_phone_number: str,
            recording_enabled: typing.Optional[bool] = None,
            outbound_greeting_recording_url: typing.Optional[str] = None,
            outbound_greeting_text: typing.Optional[str] = None,
            agent_id: typing.Optional[str] = None
    ) -> calls.Call:
        """
        Create a new call.
        
        This method is rate limited. There is no functionality to currently check for rate limits. Be careful!
        More info: https://apidocs.callrail.com/#creating-an-outbound-phone-call
        """

        logging.warning('This method is rate limited. There is no functionality to currently check for rate limits. Be careful!')

        body = {
            'caller_id': caller_id,
            'customer_phone_number': customer_phone_number,
            'business_phone_number': business_phone_number
        }

        

        if recording_enabled:
            body['recording_enabled'] = recording_enabled
        if outbound_greeting_recording_url:
            body['outbound_greeting_recording_url'] = outbound_greeting_recording_url
        if outbound_greeting_text:
            body['outbound_greeting_text'] = outbound_greeting_text
        if agent_id:
            body['agent_id'] = agent_id
        
        data = self.api_client._post(
            endpoint=f'a/{self.id}',
            path='calls.json',
            data=body
        )

        return calls.Call.from_json(
            self.api_client,
            self.id,
            typing.cast(typing.Dict[str, typing.Any], data)
        )
    
    #########################
    # Tags
    #########################

    def list_tags(
            self,
            **kwargs
    ) -> typing.Union[typing.List[tags.Tag], None]:
        """
        This endpoint returns a paginated array of tags within the target account.

        More info: https://apidocs.callrail.com/#retrieving-all-tags
        """

        pagination_type: str = kwargs.get('pagination_type', 'RELATIVE')

        sorting = kwargs.get('sorting', None)

        params = {}

        if sorting:
            params['sorting'] = sorting

        if tags_response := typing.cast(
            typing.List[typing.Dict[str, typing.Any]],
            self.api_client._get(
                endpoint=f'a/{self.id}',
                response_data_key='tags',
                path='tags.json',
                params=params or None,
                pagination_type=pagination_type,
            )
        ):
            return [tags.Tag.from_json(self.api_client, self.id, tag) for tag in tags_response]

        else:
            return None
        
    def create_tag(
            self,
            name: str,
            company_id: typing.Optional[str] = None,
            color: typing.Optional[str] = None,
            tag_level: typing.Optional[str] = None
    ) -> tags.Tag:
        """
        Create a new tag.
        
        More info: https://apidocs.callrail.com/#creating-a-tag
        """

        # some validation before sending the request
        if not name:
            raise ValueError('name is required')

        if tag_level == 'company' and not company_id:
            raise ValueError('company_id is required if tag_level is company or not "account"')

        body = {
            'name': name
        }

        if company_id:
            body['company_id'] = company_id
        if color:
            body['color'] = color
        if tag_level:
            body['tag_level'] = tag_level

        data = self.api_client._post(
            endpoint=f'a/{self.id}',
            path='tags.json',
            data=body
        )

        return tags.Tag.from_json(
            self.api_client,
            self.id,
            typing.cast(typing.Dict[str, typing.Any], data)
        )
    
    #########################
    # Companies
    #########################

    def list_companies(
            self,
            **kwargs
    ) -> typing.Union[typing.List[companies.Company], None]:
        """
        List all companies under account scope.

        More info: https://apidocs.callrail.com/#creating-a-company
        """

        pagination_type: str = kwargs.get('pagination_type', 'RELATIVE')

        sorting = kwargs.get('sorting', None)
        filtering: typing.Dict[str, typing.Any] = typing.cast(typing.Dict[str, typing.Any], kwargs.get('filtering', None))
        searching = kwargs.get('searching', None)

        if filtering:
            # only filtering field supported is status, raise a ValueError if there is a other field
            for k,v in filtering.items():
                if k != 'status':
                    raise ValueError(f'filtering field {k} is not supported')

        params = {}

        if sorting:
            params['sorting'] = sorting
        if filtering:
            params['filtering'] = filtering
        if searching:
            params['searching'] = searching

        if companies_response := typing.cast(
            typing.List[typing.Dict[str, typing.Any]],
            self.api_client._get(
                endpoint=f'a/{self.id}',
                response_data_key='companies',
                path='companies.json',
                params=params or None,
                pagination_type=pagination_type,
        )):
            return [companies.Company.from_json(self.api_client, self.id, company) for company in companies_response]
        else:
            return None
        
    def get_company(
            self,
            company_id: str,
            fields: typing.Optional[typing.MutableMapping[str, str]] = None
    ) -> companies.Company:
        """
        Get a company.
        
        More info: https://apidocs.callrail.com/#retrieving-a-single-company
        """

        if fields:
            data = self.api_client._get(
                endpoint=f'a/{self.id}',
                path=f'companies/{company_id}.json',
                params=fields
            )
        else:
            data = self.api_client._get(
                endpoint=f'a/{self.id}',
                path=f'companies/{company_id}.json'
            )

        return companies.Company.from_json(
            self.api_client,
            self.id,
            typing.cast(typing.Dict[str, typing.Any], data)
        )
    
    def create_company(
            self,
            name: str,
            time_zone: typing.Optional[str] = None
    ) -> companies.Company:
        """
        Create a new company.
        
        More info: https://apidocs.callrail.com/#creating-a-company
        """

        body = {
            'name': name
        }

        if time_zone:
            body['time_zone'] = time_zone

        data = self.api_client._post(
            endpoint=f'a/{self.id}',
            path='companies.json',
            data=body
        )

        return companies.Company.from_json(
            self.api_client,
            self.id,
            typing.cast(typing.Dict[str, typing.Any], data)
        )

    #########################
    # Form Submissions
    #########################

    def list_form_submissions(
            self,
            **kwargs
    ) -> typing.Union[typing.List[forms.FormSubmission], None]:
        """
        List form submissions.

        More information: https://apidocs.callrail.com/#listing-all-form-submissions
        """

        pagination_type: str = kwargs.get('pagination_type', 'RELATIVE')

        sorting = kwargs.get('sorting', None)
        filtering = kwargs.get('filtering', None)
        fields = kwargs.get('fields', None)

        params = {}

        if sorting:
            params['sorting'] = sorting
        if filtering:
            params['filtering'] = filtering
        if fields:
            params['fields'] = fields

        if forms_response := typing.cast(
            typing.List[typing.Dict[str, typing.Any]],
            self.api_client._get(
                endpoint=f'a/{self.id}',
                response_data_key='form_submissions',
                path='form_submissions.json',
                params=params or None,
                pagination_type=pagination_type,
            )
        ):
            return [forms.FormSubmission.from_json(self.api_client, self.id, submission) for submission in forms_response]
        else:
            logging.warning('No form submissions found')
            return None
        
    def create_form_submission(
            self,
            company_id: str,
            referrer: str,
            referring_url: str,
            landing_page_url: str,
            form_url: str,
            form_data: typing.Dict[str, typing.Any]
    ) -> forms.FormSubmission:
        """
        Create a Form Submission.
        
        More information: https://apidocs.callrail.com/#creating-a-form-submission
        """
        body = {
            'company_id': company_id,
            'referrer': referrer,
            'referring_url': referring_url,
            'landing_page_url': landing_page_url,
            'form_url': form_url,
            'form_data': form_data
        }

        data = self.api_client._post(
            endpoint=f'a/{self.id}',
            path='form_submissions.json',
            data=body
        )

        return forms.FormSubmission.from_json(
            self.api_client,
            self.id,
            typing.cast(typing.Dict[str, typing.Any], data)
        )
    
    #########################
    # Text Messages
    #########################

    def list_text_message_conversations(
            self,
            **kwargs
    ) -> typing.Union[typing.List[messages.TextMessageConversation], None]:
        """
        List all text message conversations.
        More information: https://apidocs.callrail.com/#listing-all-conversations
        """
        
        pagination_type: str = kwargs.get('pagination_type', 'RELATIVE')

        sorting = kwargs.get('sorting', None)
        filtering = kwargs.get('filtering', None)
        searching = kwargs.get('searching', None)
        fields = kwargs.get('fields', None)

        params = {}

        if sorting:
            params['sorting'] = sorting
        if filtering:
            params['filtering'] = filtering
        if searching:
            params['searching'] = searching
        if fields:
            params['fields'] = fields


        if text_messages_response := typing.cast(
            typing.List[typing.Dict[str, typing.Any]],
            self.api_client._get(
                endpoint=f'a/{self.id}',
                response_data_key='conversations',
                path='text-messages.json',
                params=params or None,
                pagination_type=pagination_type,
            )
        ):
            return [messages.TextMessageConversation.from_json(self.api_client, self.id, message) for message in text_messages_response]
        else:
            logging.warning('No text messages found')
            return None
        
    def get_text_message_conversation(
            self,
            conversation_id: str,
            fields: typing.Optional[typing.MutableMapping[str, str]] = None
    ) -> messages.TextMessageConversation:
        """
        Retrieve a single text message conversation.
        More information: https://apidocs.callrail.com/#retrieving-a-single-text-conversation
        """

        if fields:
            data = self.api_client._get(
                endpoint=f'a/{self.id}',
                path=f'text-messages/{conversation_id}.json',
                params=fields
            )
        else:
            data = self.api_client._get(
                endpoint=f'a/{self.id}',
                path=f'text-messages/{conversation_id}.json'
            )
        
        return messages.TextMessageConversation.from_json(
            self.api_client,
            self.id,
            typing.cast(typing.Dict[str, typing.Any], data)
        )
        
    def send_text(
            self,
            company_id: str,
            customer_phone_number: str,
            content: str,
            tracking_number: typing.Optional[typing.Union[str, int]] = None
    ) -> messages.TextMessageConversation:
        """
        Send a text message.
        More information: https://apidocs.callrail.com/#sending-a-text-message

        Automated messaging is strictly prohibited by CallRail. This functionality is only for
        use in person to person communication. For example, this method can be used when building a customer service
        portal initiated by a agent. 

        Predictive Data Lab is not to be held responsible for any misuse of this method.
        """
        logging.warning('Sending a text message requires approval from CallRail')
        logging.warning('Endpoint is rate limited! Be careful!')
        logging.warning('Automated messaging is not allowed. Please use this only for use in person to person communication')

        body = {
            'company_id': company_id,
            'customer_phone_number': customer_phone_number,
            'content': content
        }

        if tracking_number:
            body['tracking_number'] = tracking_number # type: ignore

        if len(content) >= 140:
            raise ValueError('Content must be less than 140 characters')
        
        data = self.api_client._post(
            endpoint=f'a/{self.id}',
            path='text-messages.json',
            data=body
        )

        return messages.TextMessageConversation.from_json(
            self.api_client,
            self.id,
            typing.cast(typing.Dict[str, typing.Any], data)
        )


import pytest
import pytest_mock
import requests_mock
import requests

from pycallrail.callrail import CallRail
from pycallrail.objects.textmessages import TextMessage, TextMessageConversation
import typing
import logging
import datetime as dt
import typeguard
import dataclasses

# Tests that a TextMessage object can be created with valid parameters. 
def test_creating_valid_text_message() -> None:
    direction = "outgoing"
    content = "Hello, this is a test message."
    created_at: dt.datetime = dt.datetime.now()
    text_message = TextMessage(direction=direction, content=content, created_at=created_at)
    assert text_message.direction == direction
    assert text_message.content == content
    assert text_message.created_at == created_at

# Tests that a TextMessage object cannot be created with empty string parameters. 
def test_empty_string_parameters() -> None:
    with pytest.raises(ValueError):
        direction = ""
        content = "Hello, this is a test message."
        created_at: dt.datetime = dt.datetime.now()
        TextMessage(direction=direction, content=content, created_at=created_at)

# Tests that a TextMessage object cannot be created with an invalid datetime format. 
def test_invalid_datetime_format() -> None:
    with pytest.raises(ValueError):
        direction = "outgpomg"
        content = "Hello, this is a test message."
        created_at = "2022-13-01 12:00:00"
        TextMessage(direction=direction, content=content, created_at=created_at)

# Tests that accessing the id attribute of a TextMessage object that has not been set returns None. 
def test_accessing_unset_id_attribute() -> None:
    text_message = TextMessage(
        direction="outgoing", 
        content="Hello, this is a test message.", 
        created_at=dt.datetime.now())
    assert text_message.id is None

# Tests archiving a TextMessageConversation. 
def test_archive_conversation(mocker: pytest_mock.MockerFixture) -> None:
    # Happy path test for archiving a TextMessageConversation
    api_client = CallRail('test_key')
    account_id = 'test_account'
    conversation_id = 'test_conversation'
    conversation_data: typing.Dict[str, str] = {
        'id': conversation_id,
        'state': 'active'
    }
    conversation = TextMessageConversation(api_client, account_id, **conversation_data)

    # Mock the _put method of the api_client to simulate a successful archive request
    mocker.patch.object(api_client, '_put', return_value=None)

    conversation.archive()

    assert conversation.state == 'archived'

# Tests creating a TextMessageConversation object with valid arguments. 
def test_create_text_message_conversation_valid_args() -> None:
    # Happy path test for creating a TextMessageConversation with valid arguments
    api_client = CallRail('test_key')
    account_id = 'test_account'
    conversation_data: typing.Dict[str, typing.Any] = {
        'id': 'test_conversation',
        'company_id': 'test_company',
        'initial_tracker_id': 'test_initial_tracker',
        'current_tracker_id': 'test_current_tracker',
        'customer_name': 'John Doe',
        'customer_phone_number': '+1234567890',
        'initial_tracking_number': '+0987654321',
        'current_tracking_number': '+1234567890',
        'last_message_at': dt.datetime.now(),
        'state': 'active',
        'company_time_zone': 'UTC',
        'formatted_customer_phone_number': '(123) 456-7890',
        'formatted_initial_tracking_number': '(098) 765-4321',
        'formatted_current_tracking_number': '(123) 456-7890',
        'formatted_customer_name': 'Doe, John',
        'recent_messages': []
    }

    conversation = TextMessageConversation(api_client, account_id, **conversation_data)

    assert conversation.id == conversation_data['id']
    assert conversation.customer_name == conversation_data['customer_name']
    assert conversation.recent_messages == conversation_data['recent_messages']


# Tests deserializing JSON data into a TextMessageConversation object with missing or invalid fields. 
def test_deserialize_json_missing_fields() -> None:
    # Edge case test for deserializing JSON data into a TextMessageConversation with missing or invalid fields
    api_client = CallRail('test_key')
    account_id = 'test_account'
    json_data = {
        # Missing id field
        'company_id': 'test_company',
        'initial_tracker_id': 'test_initial_tracker',
        'current_tracker_id': 'test_current_tracker',
        'customer_name': 'John Doe',
        'customer_phone_number': '+1234567890',
        'initial_tracking_number': '+0987654321',
        'current_tracking_number': '+1234567890',
        'last_message_at': dt.datetime.now().isoformat(),
        'state': 'active',
        'company_time_zone': 'UTC',
        'formatted_customer_phone_number': '(123) 456-7890',
        'formatted_initial_tracking_number': '(098) 765-4321',
        'formatted_current_tracking_number': '(123) 456-7890',
        'formatted_customer_name': 'Doe, John',
        'recent_messages': []
    }

    with pytest.raises(KeyError):
        TextMessageConversation.from_json(api_client, account_id, json_data)

# Tests deserializing JSON data into a TextMessageConversation object with valid data. 
def test_deserialize_json_valid_data() -> None:
    # Happy path test for deserializing JSON data into a TextMessageConversation with valid data
    api_client = CallRail('test_key')
    account_id = 'test_account'
    json_data = {
        'id': 'test_conversation',
        'company_id': 'test_company',
        'initial_tracker_id': 'test_initial_tracker',
        'current_tracker_id': 'test_current_tracker',
        'customer_name': 'John Doe',
        'customer_phone_number': '+1234567890',
        'initial_tracking_number': '+0987654321',
        'current_tracking_number': '+1234567890',
        'last_message_at': '2022-01-01T00:00:00Z',
        'state': 'active',
        'company_time_zone': 'UTC',
        'formatted_customer_phone_number': '(123) 456-7890',
        'formatted_initial_tracking_number': '(098) 765-4321',
        'formatted_current_tracking_number': '(123) 456-7890',
        'formatted_customer_name': 'Doe, John',
        'recent_messages': [
            {
                'direction': 'incoming',
                'content': 'Hello',
                'created_at': '2022-01-01T00:00:00Z'
            },
            {
                'direction': 'outgoing',
                'content': 'Hi there',
                'created_at': '2022-01-01T00:01:00Z'
            }
        ]
    }

    conversation = TextMessageConversation.from_json(api_client, account_id, json_data)

    assert conversation.id == json_data['id']
    assert conversation.customer_name == json_data['customer_name']
    assert len(conversation.recent_messages) == len(json_data['recent_messages'])

# Tests accessing attributes of a TextMessageConversation object. 
def test_access_attributes() -> None:
    # Happy path test for accessing attributes of a TextMessageConversation object
    api_client = CallRail('test_key')
    account_id = 'test_account'
    conversation_data = {
        'id': 'test_conversation',
        'company_id': 'test_company',
        'initial_tracker_id': 'test_initial_tracker',
        'current_tracker_id': 'test_current_tracker',
        'customer_name': 'John Doe',
        'customer_phone_number': '+1234567890',
        'initial_tracking_number': '+0987654321',
        'current_tracking_number': '+1234567890',
        'last_message_at': dt.datetime.now(),
        'state': 'active',
        'company_time_zone': 'UTC',
        'formatted_customer_phone_number': '(123) 456-7890',
        'formatted_initial_tracking_number': '(098) 765-4321',
        'formatted_current_tracking_number': '(123) 456-7890',
        'formatted_customer_name': 'Doe, John',
        'recent_messages': []
    }

    conversation = TextMessageConversation(api_client, account_id, **conversation_data)

    assert conversation.id == conversation_data['id']
    assert conversation.customer_name == conversation_data['customer_name']
    assert conversation.recent_messages == conversation_data['recent_messages']